"""Utilities for creating and storing figures."""
import os.path
import statistics as st
import warnings
from math import isfinite, sqrt
from typing import Callable, Final, Iterable, Sequence, cast

import matplotlib.pyplot  # type: ignore
from matplotlib.artist import Artist  # type: ignore
from matplotlib.axes import Axes  # type: ignore
from matplotlib.backend_bases import RendererBase  # type: ignore
from matplotlib.backends.backend_agg import RendererAgg  # type: ignore
from matplotlib.figure import Figure, SubplotBase  # type: ignore

import moptipy.utils.plot_defaults as pd
from moptipy.utils.lang import Lang
from moptipy.utils.path import Path
from moptipy.utils.types import type_error, type_name_of

#: the golden ratio
__GOLDEN_RATIO: Final[float] = 0.5 + (0.5 * sqrt(5))


def create_figure(width: float | int | None = 8.6,
                  height: float | int | None = None,
                  dpi: float | int | None = 384.0,
                  **kwargs) -> Figure:
    """
    Create a matplotlib figure.

    :param width: the optional width
    :param height: the optional height
    :param dpi: the dpi value
    :param kwargs: a set of optional arguments
    :return: the figure option
    """
    # Check the size, i.e., width and height and use
    put_size: bool = True
    if width is None:
        if height is None:
            put_size = False
        else:
            width = __GOLDEN_RATIO * height
    elif height is None:
        height = width / __GOLDEN_RATIO

    if put_size:
        if isinstance(height, int):
            height = float(height)
        if not isinstance(height, float):
            raise ValueError(f"Invalid height type {type(height)}.")
        if not (isfinite(height) and (0.1 < height < 10000.0)):
            raise ValueError(f"Invalid height {height}.")
        nheight = int(0.5 + (height * 72)) / 72.0
        if not (isfinite(nheight) and (0.1 < nheight < 10000.0)):
            raise ValueError(f"Invalid height {height} as it maps to "
                             f"{nheight} after to-point-and-back conversion.")

        if isinstance(width, int):
            width = float(width)
        if not isinstance(width, float):
            raise ValueError(f"Invalid width type {type(width)}.")
        if not (isfinite(width) and (0.1 < width < 10000.0)):
            raise ValueError(f"Invalid width {width}.")
        nwidth = int(0.5 + (width * 72)) / 72.0
        if not (isfinite(nwidth) and (0.1 < nwidth < 10000.0)):
            raise ValueError(f"Invalid width {width} as it maps to "
                             f"{nwidth} after to-point-and-back conversion.")

        kwargs["figsize"] = width, height

    if dpi is not None:
        if isinstance(dpi, int):
            dpi = float(dpi)
        if not isinstance(dpi, float):
            raise ValueError(f"Invalid dpi type {type(dpi)}.")
        if not (isfinite(dpi) and (1.0 < dpi < 10000.0)):
            raise ValueError(f"Invalid dpi value {dpi}.")
        kwargs["dpi"] = dpi

    if "frameon" not in kwargs:
        kwargs["frameon"] = False

    if "constrained_layout" not in kwargs:
        kwargs["constrained_layout"] = False

    return Figure(**kwargs)


def __divide_evenly(items: int, chunks: int, reverse: bool) -> list[int]:
    """
    Divide `n` items into `k` chunks, trying to create equally-sized chunks.

    :param items: the number of items to divide
    :param chunks: the number of chunks
    :param reverse: should we put the additional items at the end (True)
        or front (False)
    :returns: the list of items

    >>> print(__divide_evenly(9, 3, reverse=True))
    [3, 3, 3]
    >>> print(__divide_evenly(10, 3, reverse=True))
    [3, 3, 4]
    >>> print(__divide_evenly(10, 3, reverse=False))
    [4, 3, 3]
    >>> print(__divide_evenly(11, 3, reverse=True))
    [3, 4, 4]
    >>> print(__divide_evenly(11, 3, reverse=False))
    [4, 4, 3]
    >>> print(__divide_evenly(12, 3, reverse=False))
    [4, 4, 4]
    """
    # validate that we do not have more chunks than items or otherwise invalid
    # parameters
    if (items <= 0) or (chunks <= 0) or (chunks > items):
        raise ValueError(f"cannot divide {items} items into {chunks} chunks.")

    # First we compute the minimum number of items per chunk.
    # Our basic solution is to put exactly this many items into each chunk.
    # For example, if items=10 and chunks=3, this will fill 3 items into each
    # chunk.
    result: list[int] = [items // chunks] * chunks
    # We then fill the remaining items into the chunks at the front
    # (reverse=False) or end (reverse=True) of the list.
    # We do this by putting exactly one such item into each chunk, starting
    # from the end.
    # Since items modulo chunks must be in [0..chunks), this is possible.
    # This setup will yield the lowest possible standard deviation of items
    # per chunk.
    # In case of items=10 and chunks=3, items % chunks = 1 and the chunks will
    # become [3, 3, 4]. If items=3, no items need to be distributed and we get
    # [3, 3, 3].
    if reverse:
        for i in range(1, (items % chunks) + 1):
            result[-i] += 1
    else:
        for i in range(items % chunks):
            result[i] += 1
    return result


def create_figure_with_subplots(
        items: int,
        max_items_per_plot: int = 3,
        max_rows: int = 3,
        max_cols: int = 1,
        min_rows: int = 1,
        min_cols: int = 1,
        default_width_per_col: float | int | None = 8.6,
        max_width: float | int | None = 8.6,
        default_height_per_row: float | int | None = 8.6 / __GOLDEN_RATIO,
        max_height: float | int | None = 9,
        dpi: float | int | None = 384.0,
        **kwargs) \
        -> tuple[Figure, tuple[tuple[SubplotBase | Figure,
                                     int, int, int, int, int], ...]]:
    """
    Divide a figure into nrows*ncols sub-plots.

    :param items: the number of items to divide
    :param max_items_per_plot: the maximum number of items per plot
    :param max_rows: the maximum number of rows
    :param max_cols: the maximum number of columns
    :param min_rows: the minimum number of rows
    :param min_cols: the minimum number of cols
    :param default_width_per_col: the optional default width of a column
    :param default_height_per_row: the optional default height per row
    :param max_height: the maximum height
    :param max_width: the maximum width
    :param dpi: the dpi value
    :param kwargs: a set of optional arguments
    :returns: a tuple with the figure, followed by a series of tuples with
        each sub-figure, the index of the first item assigned to it, the
        index of the last item assigned to it + 1, their row, their column,
        and their overall index
    """
    # First, we do a lot of sanity checks
    if not isinstance(items, int):
        raise type_error(items, "items", int)
    if items < 1:
        raise ValueError(f"items must be positive, but is {items}.")
    if not isinstance(max_items_per_plot, int):
        raise type_error(max_items_per_plot, "max_items_per_plot", int)
    if max_items_per_plot < 1:
        raise ValueError(f"max_items_per_plot must be positive, "
                         f"but is {max_items_per_plot}.")
    if not isinstance(max_rows, int):
        raise type_error(max_rows, "max_rows", int)
    if max_rows < 1:
        raise ValueError(f"max_rows must be positive, but is {max_rows}.")
    if not isinstance(min_rows, int):
        raise type_error(min_rows, "min_rows", int)
    if min_rows < 1:
        raise ValueError(f"min_rows must be positive, but is {min_rows}.")
    if min_rows > max_rows:
        raise ValueError(
            f"min_rows ({min_rows}) must be <= max_rows ({max_rows}).")
    if not isinstance(max_cols, int):
        raise type_error(max_cols, "max_cols", int)
    if max_cols < 1:
        raise ValueError(f"max_cols must be positive, but is {max_cols}.")
    if not isinstance(min_cols, int):
        raise type_error(min_cols, "min_cols", int)
    if min_cols < 1:
        raise ValueError(f"min_cols must be positive, but is {min_cols}.")
    if min_cols > max_cols:
        raise ValueError(
            f"min_cols ({min_cols}) must be <= max_cols ({max_cols}).")
    if (max_cols * max_rows * max_items_per_plot) < items:
        raise ValueError(
            f"Cannot distribute {items} items into at most {max_rows} rows "
            f"and {max_cols} cols with at most {max_items_per_plot} per "
            "plot.")
    if default_width_per_col is not None:
        default_width_per_col = float(default_width_per_col)
        if (not isfinite(default_width_per_col)) \
                or (default_width_per_col <= 0.1) \
                or (default_width_per_col >= 1000):
            raise ValueError(
                f"invalid default_width_per_col {default_width_per_col}")
    if max_width is not None:
        max_width = float(max_width)
        if (not isfinite(max_width)) \
                or (max_width <= 0.1) or (max_width >= 10000):
            raise ValueError(f"invalid max_width {max_width}")
        if (default_width_per_col is not None) \
                and (default_width_per_col > max_width):
            raise ValueError(f"default_width_per_col {default_width_per_col} "
                             f"> max_width {max_width}")
    if default_height_per_row is not None:
        default_height_per_row = float(default_height_per_row)
        if (not isfinite(default_height_per_row)) \
                or (default_height_per_row <= 0.1) \
                or (default_height_per_row >= 10000):
            raise ValueError(
                f"invalid default_height_per_row {default_height_per_row}")
    if max_height is not None:
        max_height = float(max_height)
        if (not isfinite(max_height)) \
                or (max_height <= 0.1) or (max_height >= 10000):
            raise ValueError(f"invalid max_width {max_height}")
        if (default_height_per_row is not None) \
                and (max_height < default_height_per_row):
            raise ValueError(
                f"max_height {max_height} < "
                f"default_height_per_row {default_height_per_row}")

    # setting up the default dimensions
    if default_height_per_row is None:
        default_height_per_row = 1.95
        if max_height is not None:
            default_height_per_row = min(default_height_per_row, max_height)
    if max_height is None:
        max_height = max_rows * default_height_per_row
    if default_width_per_col is None:
        default_width_per_col = 8.6
        if max_width is not None:
            default_width_per_col = min(default_width_per_col, max_width)
    if max_width is None:
        max_width = default_width_per_col

    # How many plots do we need at least? Well, if we can put
    # max_items_per_plot items into each plot and have items many items, then
    # we need ceil(items / max_items_per_plot) many. This is what we compute
    # here.
    min_plots: Final[int] = (items // max_items_per_plot) \
        + min(1, items % max_items_per_plot)
    # The maximum conceivable number of plots would be items, so this we do
    # not need to compute.
    plots_i: Final[int] = 6
    rows_i: Final[int] = plots_i + 1
    cols_i: Final[int] = rows_i + 1
    height_i: Final[int] = cols_i + 1
    width_i: Final[int] = height_i + 1
    plots_per_row_i: Final[int] = width_i + 1
    chunks_i: Final[int] = plots_per_row_i + 1
    best: tuple[float,  # the plots-per-row std
                float,  # the items-per-plot std
                float,  # the overall area of the figure
                float,  # the std deviation of (w, h*golden_ratio),
                float,  # -plot area
                float,  # the std deviation of (w, h*golden_ratio) per plot
                int,  # the number of plots
                int,  # the number of rows
                int,  # the number of cols
                float,  # the figure height
                float,  # the figure width
                list[int],  # the plots per row
                list[int]] = \
        (1000000.0, 1000000.0, 1000000.0, 1000000.0, 1000000.0, 1000000.0,
         1 << 62, 1 << 62, 1 << 62, 1000000.0, 1000000.0, [0], [0])

    # Now we simply try all valid combinations.
    for rows in range(min_rows, max_rows + 1):
        if rows > items:
            continue  # more rows than items
        for cols in range(min_cols, max_cols + 1):
            # We need cols plots per row, except for the last row, where we
            # could only put a single plot if we wanted.
            min_plots_for_config: int = 1 + ((rows - 1) * cols)
            if min_plots_for_config > items:
                continue  # if this exceeds the maximum number of plots, skip
            # compute the maximum plots we can get in this config
            max_plots_per_config: int = rows * cols
            if max_plots_per_config < min_plots:
                continue  # not enough plots possible

            # now we can iterate over the number of plots we can provide
            for plots in range(max(min_plots_for_config, min_plots),
                               min(max_plots_per_config, items) + 1):
                # We compute the figure width. If we have multiple columns,
                # we try to add 2% plus 0.02" as breathing space per column.
                fig_width = min(max_width, ((1.02 * default_width_per_col)
                                            + 0.02) * cols)
                # The width of a single plot is then computed as follows:
                plot_width = fig_width / cols
                if plot_width < 0.9:
                    continue  # the single plots would be too small
                # We compute the overall figure height. Again we add breathing
                # space if we have multiple rows.
                fig_height = min(max_height, ((1.02 * default_height_per_row)
                                              + 0.02) * rows)
                plot_height = fig_height / rows
                if plot_height < 0.9:
                    continue  # the single plots would be too small
                # So dimension-wise, the plot sizes are OK.
                # How about the distribution of items? We put the rows with
                # plots at the top and the plots with more items at the end.
                plot_distr = __divide_evenly(plots, rows, reverse=False)
                item_distr = __divide_evenly(items, plots, reverse=True)

                current: tuple[float, float, float, float, float, float,
                               int, int, int, float, float,
                               list[int], list[int]] = (
                    st.stdev(plot_distr) if rows > 1 else 0,
                    st.stdev(item_distr) if plots > 1 else 0,
                    fig_height * fig_width,  # the area of the figure
                    st.stdev([fig_width, fig_height * __GOLDEN_RATIO]),
                    -plot_width * plot_height,  # -plot area
                    st.stdev([plot_width, plot_height * __GOLDEN_RATIO]),
                    plots,  # the number of plots
                    rows,  # the number of rows
                    cols,  # the number of cols
                    fig_height,  # the figure height
                    fig_width,  # the figure width
                    plot_distr,  # the plots per row
                    item_distr)  # the data chunks per plot
                if ((current[plots_i] + 1)  # type: ignore
                    < best[plots_i]) or ((  # type: ignore
                        current[plots_i]   # type: ignore
                        <= (best[plots_i] + 1))  # type: ignore
                        and (current < best)):  # type: ignore
                    best = current

    n_plots: Final[int] = best[plots_i]  # type: ignore
    if n_plots > items:
        raise ValueError(
            f"Could not place {items} in {min_cols}..{max_cols} columns "
            f"at {min_rows}..{max_rows} rows with at most "
            f"{max_items_per_plot} items per plot for a max_width={max_width}"
            f" and max_height={max_height}.")

    # create the figure of the computed dimensions
    figure: Final[Figure] = create_figure(
        width=best[width_i], height=best[height_i],  # type: ignore
        dpi=dpi, **kwargs)
    if n_plots <= 1:  # if there is only one plot, we are done here
        return figure, ((figure, 0, items, 0, 0, 0), )

    # if there are multiple plots, we need to generate them
    allfigs: list[tuple[SubplotBase | Figure,
                        int, int, int, int, int]] = []
    index: int = 0
    chunk_start: int = 0
    nrows: Final[int] = best[rows_i]  # type: ignore
    ncols: Final[int] = best[cols_i]  # type: ignore
    chunks: Final[list[int]] = best[chunks_i]  # type: ignore
    plots_per_row: Final[list[int]] = best[plots_per_row_i]  # type: ignore
    for i in range(nrows):
        for j in range(plots_per_row[i]):
            chunk_next = chunk_start + chunks[index]
            allfigs.append((figure.add_subplot(nrows, ncols,
                                               (i * ncols) + j + 1),
                            chunk_start, chunk_next, i, j, index))
            chunk_start = chunk_next
            index += 1
    return figure, tuple(allfigs)


def save_figure(fig: Figure,
                file_name: str = "figure",
                dir_name: str = ".",
                formats: str | Iterable[str] = "svg") -> list[Path]:
    """
    Store the given figure in files of the given formats and dispose it.

    :param fig: the figure to save
    :param file_name: the basic file name
    :param dir_name: the directory name
    :param formats: the file format(s)
    :return: a list of files
    """
    if not isinstance(fig, Figure):
        raise type_error(fig, "figure", Figure)
    if not isinstance(file_name, str):
        raise type_error(file_name, "file_name", str)
    if len(file_name) <= 0:
        raise ValueError(f"Invalid filename '{file_name}'.")
    if not isinstance(dir_name, str):
        raise type_error(dir_name, "dir_name", str)
    if len(dir_name) <= 0:
        raise ValueError(f"Invalid dirname '{dir_name}'.")
    if isinstance(formats, str):
        formats = [formats]
    if not isinstance(formats, Iterable):
        raise type_error(formats, "formats", Iterable)

    size = fig.get_size_inches()
    if size[0] >= size[1]:
        orientation = "landscape"
    else:
        orientation = "portrait"

    # set minimal margins to the axes to avoid wasting space
    for ax in fig.axes:
        ax.margins(0, 0)

    use_dir = Path.directory(dir_name)
    files = []
    for fmt in formats:
        if not isinstance(fmt, str):
            raise type_error(fmt, "element of formats", str)
        dest_file = Path.path(os.path.join(use_dir, f"{file_name}.{fmt}"))
        dest_file.ensure_file_exists()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            # UserWarning: There are no gridspecs with layoutgrids.
            # Possibly did not call parent GridSpec with the "figure" keyword
            fig.savefig(dest_file, transparent=True, format=fmt,
                        orientation=orientation,
                        dpi="figure",
                        bbox_inches="tight",
                        pad_inches=1.0 / 72.0)
        dest_file.enforce_file()
        files.append(dest_file)

    fig.clf(False)
    matplotlib.pyplot.close(fig)
    del fig

    if len(files) <= 0:
        raise ValueError("No formats were specified.")

    return files


def label_box(axes: Axes,
              text: str,
              x: float | None = None,
              y: float | None = None,
              font_size: float = pd.importance_to_font_size(0),
              may_rotate_text: bool = False,
              z_order: float | None = None,
              font: None | str | Callable =
              lambda: Lang.current().font()) -> None:
    """
    Put a label text box near an axis.

    :param axes: the axes to add the label to
    :param text: the text to place
    :param x: the location along the x-axis: `0` means left,
        `0.5` means centered, `1` means right
    :param y: the location along the x-axis: `0` means bottom,
        `0.5` means centered, `1` means top
    :param font_size: the font size
    :param may_rotate_text: should we rotate the text by 90° if that
        makes sense (`True`) or always keep it horizontally (`False`)
    :param z_order: an optional z-order value
    :param font: the font to use
    """
    if x is None:
        if y is None:
            raise ValueError("At least one of x or y must not be None.")
        x = 0
    elif y is None:
        y = 0

    spacing: Final[float] = max(4.0, font_size / 2.0)
    xtext: float = 0.0
    ytext: float = 0.0
    xalign: str = "center"
    yalign: str = "center"
    if x >= 0.85:
        xtext = -spacing
        xalign = "right"
    elif x <= 0.15:
        xtext = spacing
        xalign = "left"
    if y >= 0.85:
        ytext = -spacing
        yalign = "top"
    elif y <= 0.15:
        ytext = spacing
        yalign = "bottom"

    args = {"text": text,
            "xy": (x, y),
            "xytext": (xtext, ytext),
            "verticalalignment": yalign,
            "horizontalalignment": xalign,
            "xycoords": "axes fraction",
            "textcoords": "offset points",
            "fontsize": font_size,
            "bbox": {"boxstyle": "round",
                     "color": "white",
                     "fill": True,
                     "linewidth": 0,
                     "alpha": 0.9}}
    if z_order is not None:
        args["zorder"] = z_order

    if may_rotate_text and (len(text) > 2):
        args["rotation"] = 90

    if callable(font):
        font = font()
    if font is not None:
        if not isinstance(font, str):
            raise type_error(font, "font", str)
        args["fontname"] = font

    axes.annotate(**args)


def label_axes(axes: Axes,
               x_label: str | None = None,
               x_label_inside: bool = True,
               x_label_location: float = 0.5,
               y_label: str | None = None,
               y_label_inside: bool = True,
               y_label_location: float = 1,
               font_size: float = pd.importance_to_font_size(0),
               z_order: float | None = None) -> None:
    """
    Put labels on a figure.

    :param axes: the axes to add the label to
    :param x_label: a callable returning the label for
        the x-axis, a label string, or `None` if no label should be put
    :param x_label_inside: put the x-axis label inside the plot (so that
        it does not consume additional vertical space)
    :param x_label_location: the location of the x-axis label if it is
        placed inside the plot area
    :param y_label: a callable returning the label for
        the y-axis, a label string, or `None` if no label should be put
    :param y_label_inside: put the xyaxis label inside the plot (so that
        it does not consume additional horizontal space)nal vertical space)
    :param y_label_location: the location of the y-axis label if it is
        placed inside the plot area
    :param font_size: the font size to use
    :param z_order: an optional z-order value
    """
    # put the label on the x-axis, if any
    if x_label is not None:
        if not isinstance(x_label, str):
            raise type_error(x_label, "x_label", str)
        if len(x_label) > 0:
            if x_label_inside:
                label_box(axes, text=x_label, x=x_label_location, y=0,
                          font_size=font_size, z_order=z_order)
            else:
                axes.set_xlabel(x_label, fontsize=font_size)

    # put the label on the y-axis, if any
    if y_label is not None:
        if not isinstance(y_label, str):
            raise type_error(y_label, "y_label", str)
        if len(y_label) > 0:
            if y_label_inside:
                label_box(axes, text=y_label, x=0, y=y_label_location,
                          font_size=font_size, may_rotate_text=True,
                          z_order=z_order)
            else:
                axes.set_ylabel(y_label, fontsize=font_size)


def get_axes(figure: Axes | SubplotBase | Figure) -> Axes:
    """
    Obtain the axes from a figure or axes object.

    :param figure: the figure
    :return: the Axes
    """
    if isinstance(figure, Figure):
        return figure.add_axes([0.005, 0.005, 0.99, 0.99])
    if hasattr(figure, "axes") \
            or isinstance(getattr(type(figure), "axes", None), property):
        try:
            if isinstance(figure.axes, Axes):
                return cast(Axes, figure.axes)
            if isinstance(figure.axes, Sequence):
                ax = figure.axes[0]
                if isinstance(ax, Axes):
                    return cast(Axes, ax)
            elif isinstance(figure.axes, Iterable):
                for k in figure.axes:
                    if isinstance(k, Axes):
                        return cast(Axes, k)
                    break
        except TypeError:
            pass
        except IndexError:
            pass
    if isinstance(figure, Axes):
        return cast(Axes, figure)
    raise TypeError(
        f"Cannot get Axes of object of type {type_name_of(figure)}.")


def get_renderer(figure: SubplotBase | Axes | Figure) -> RendererBase:
    """
    Get a renderer that can be used for determining figure element sizes.

    :param figure: the figure element
    :return: the renderer
    """
    if isinstance(figure, (Axes, SubplotBase)):
        figure = figure.figure
    if not isinstance(figure, Figure):
        raise type_error(figure, "figure", Figure)
    canvas = figure.canvas
    if hasattr(canvas, "renderer"):
        return canvas.renderer
    if hasattr(canvas, "get_renderer"):
        return canvas.get_renderer()
    return RendererAgg(width=figure.get_figwidth(),
                       height=figure.get_figheight(),
                       dpi=figure.get_dpi())


def cm_to_inch(cm: int | float) -> float:
    """
    Convert cm to inch.

    :param cm: the cm value
    :return: the value in inch
    """
    if not isinstance(cm, int):
        if not isinstance(cm, float):
            raise type_error(cm, "cm", (int, float))
        if not isfinite(cm):
            raise ValueError(f"cm must be finite, but is {cm}.")
    res: float = cm / 2.54
    if not isfinite(res):
        raise ValueError(f"Conversation {cm} cm to inch "
                         f"must be finite, but is {res}.")
    return res


#: the color attributes
__COLOR_ATTRS: Final[tuple[tuple[bool, bool, str], ...]] = \
    ((True, True, "get_label"), (False, True, "label"),
     (False, True, "_label"), (True, False, "get_color"),
     (False, False, "color"), (False, False, "_color"),
     (False, False, "edgecolor"), (False, False, "_edgecolor"),
     (False, False, "markeredgecolor"), (False, False, "_markeredgecolor"))


def get_label_colors(
        handles: Iterable[Artist],
        color_map: dict[str, tuple[float, ...] | str] | None = None,
        default_color: tuple[float, ...] | str = pd.COLOR_BLACK) \
        -> list[tuple[float, ...] | str]:
    """
    Get a list with label colors from a set of artists.

    :param handles: the handles
    :param color_map: an optional color decode
    :param default_color: the default color
    :returns: a list of label colors
    """
    if not isinstance(handles, Iterable):
        raise type_error(handles, "handles", Iterable)

    def __get_color(a: Artist, colmap=color_map,
                    defcol=default_color) -> tuple[float, ...] | str:
        if not isinstance(a, Artist):
            raise type_error(a, "artist", Artist)

        for acall, astr, aname in __COLOR_ATTRS:
            if hasattr(a, aname):
                val = getattr(a, aname)
                if val is None:
                    continue
                if acall:
                    val = val()
                if val is None:
                    continue
                if astr:
                    if colmap:
                        if val in colmap:
                            val = colmap[val]
                        else:
                            continue
                    else:
                        continue
                if val is None:
                    continue
                if val != defcol:
                    return val
        return defcol

    return [__get_color(aa) for aa in handles]



"""Test stuff on real vectors."""
from math import exp, inf, isfinite
from typing import Any, Callable, Final, Iterable

import numpy as np
from numpy.random import Generator, default_rng

from moptipy.api.algorithm import Algorithm
from moptipy.api.objective import Objective
from moptipy.api.operators import Op0
from moptipy.examples.vectors.ackley import Ackley
from moptipy.spaces.vectorspace import VectorSpace
from moptipy.tests.algorithm import validate_algorithm
from moptipy.tests.op0 import validate_op0
from moptipy.utils.nputils import DEFAULT_FLOAT
from moptipy.utils.types import type_error

#: The dimensions for tests
DIMENSIONS_FOR_TESTS: Final[tuple[int, ...]] = (1, 2, 3, 4, 5, 10)


def __lbub(random: Generator) -> tuple[float, float]:
    """
    Generate a pair of lower- and upper bounds.

    :param random: the random number generator
    :returns: a tuple with the lower and upper bound
    """
    while True:
        lb = inf
        ub = inf
        while not isfinite(lb):
            lb = exp(20.0 * random.normal())

        i = random.integers(3)
        if i == 0:
            lb = 0.0
        elif i == 11:
            lb = -lb

        while not isfinite(ub):
            ub = exp(20.0 * random.normal())
        i = random.integers(3)
        if i == 0:
            ub = 0.0
        elif i == 11:
            ub = -ub

        if ub < lb:
            lb, ub = ub, lb

        df = ub - lb
        if isfinite(df) and (df > 1e-9):
            return lb, ub


def vectors_for_tests(dims: Iterable[int] = DIMENSIONS_FOR_TESTS) \
        -> Iterable[VectorSpace]:
    """
    Get a sequence of vector spaces for tests.

    :param dims: the dimensions
    :returns: the sequence of vector spaces
    """
    if not isinstance(dims, Iterable):
        raise type_error(dims, "dims", Iterable)

    random: Final[Generator] = default_rng()
    spaces: Final[list[VectorSpace]] = []
    for idx, dim in enumerate(dims):
        if not isinstance(dim, int):
            raise type_error(dim, f"dims[{idx}]", int)
        if dim <= 0:
            raise ValueError(f"dims[{idx}]={dim}")

        # allocate bounds arrays
        lbv: np.ndarray = np.empty(dim, DEFAULT_FLOAT)
        ubv: np.ndarray = np.empty(dim, DEFAULT_FLOAT)
        for i in range(dim):  # fill bound arrays
            lbv[i], ubv[i] = __lbub(random)

        spaces.append(VectorSpace(
            dim, lbv if random.integers(2) <= 0 else float(min(lbv)),
            ubv if random.integers(2) <= 0 else float(max(ubv))))

    if 1 in dims:
        spaces.append(VectorSpace(1, -1.0, 1.0))
    if 2 in dims:
        spaces.append(VectorSpace(2, 0.0, 1.0))
    if 3 in dims:
        spaces.append(VectorSpace(3, -1.0, 0.0))
    return tuple(spaces)


def validate_algorithm_on_vectors(
        objective: Objective | Callable[[VectorSpace], Objective],
        algorithm: Algorithm | Callable[[VectorSpace, Objective], Algorithm],
        max_fes: int = 100,
        uses_all_fes_if_goal_not_reached=True,
        dims: Iterable[int] = DIMENSIONS_FOR_TESTS) -> None:
    """
    Check the validity of a black-box algorithm on vector problems.

    :param algorithm: the algorithm or algorithm factory
    :param objective: the objective function or function factory
    :param max_fes: the maximum number of FEs
    :param uses_all_fes_if_goal_not_reached: will the algorithm use all FEs
        unless it reaches the goal?
    :param dims: the dimensions
    """
    if not (isinstance(algorithm, Algorithm) or callable(algorithm)):
        raise type_error(algorithm, "algorithm", Algorithm, True)
    if not (isinstance(objective, Objective) or callable(objective)):
        raise type_error(objective, "objective", Objective, True)
    if not isinstance(dims, Iterable):
        raise type_error(dims, "dims", Iterable)

    for space in vectors_for_tests(dims):
        if callable(objective):
            objf = objective(space)
            if not isinstance(objf, Objective):
                raise type_error(objf, "result of callable objective",
                                 Objective)
        else:
            objf = objective
        if callable(algorithm):
            algo = algorithm(space, objf)
            if not isinstance(algo, Algorithm):
                raise type_error(algo, "result of callable algorithm",
                                 Algorithm)
        else:
            algo = algorithm

        validate_algorithm(
            algorithm=algo, solution_space=space, objective=objf,
            max_fes=max_fes,
            uses_all_fes_if_goal_not_reached=uses_all_fes_if_goal_not_reached)


def make_vector_valid(space: VectorSpace) -> \
        Callable[[Generator, np.ndarray], np.ndarray]:
    """
    Create a function that can make a vector space element valid.

    :param space: the vector space
    :returns: the function
    """

    def __make_valid(prnd: Generator,
                     x: np.ndarray,
                     ppp=space) -> np.ndarray:
        np.copyto(x, prnd.uniform(ppp.lower_bound,
                                  ppp.upper_bound, ppp.dimension))
        return x

    return __make_valid


def validate_algorithm_on_ackley(
        algorithm: Algorithm | Callable[[VectorSpace, Objective], Algorithm],
        uses_all_fes_if_goal_not_reached: bool = True,
        dims: Iterable[int] = DIMENSIONS_FOR_TESTS) -> None:
    """
    Check the validity of a black-box algorithm on Ackley's function.

    :param algorithm: the algorithm or algorithm factory
    :param uses_all_fes_if_goal_not_reached: will the algorithm use all FEs
        unless it reaches the goal?
    :param dims: the dimensions
    """
    validate_algorithm_on_vectors(
        Ackley(), algorithm,
        uses_all_fes_if_goal_not_reached=uses_all_fes_if_goal_not_reached,
        dims=dims)


def validate_op0_on_1_vectors(
        op0: Op0 | Callable[[VectorSpace], Op0],
        search_space: VectorSpace,
        number_of_samples: int | None = None,
        min_unique_samples:
        int | Callable[[int, VectorSpace], int] | None
        = lambda i, _: max(1, i // 3)) -> None:
    """
    Validate the nullary operator on one `VectorSpace` instance.

    :param op0: the operator or operator factory
    :param search_space: the search space
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    """
    args: dict[str, Any] = {
        "op0": op0(search_space) if callable(op0) else op0,
        "search_space": search_space,
        "make_search_space_element_valid":
            make_vector_valid(search_space)
    }
    if number_of_samples is not None:
        args["number_of_samples"] = number_of_samples
    if min_unique_samples is not None:
        args["min_unique_samples"] = min_unique_samples
    validate_op0(**args)


def validate_op0_on_vectors(
        op0: Op0 | Callable[[VectorSpace], Op0],
        number_of_samples: int | None = None,
        min_unique_samples:
        int | Callable[[int, VectorSpace], int] | None
        = lambda i, _: max(1, i // 3)) -> None:
    """
    Validate the nullary operator on default `VectorSpace` instance.

    :param op0: the operator or operator factory
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    """
    for vs in vectors_for_tests():
        validate_op0_on_1_vectors(
            op0, vs, number_of_samples, min_unique_samples)



"""A nullary operator forwarding to another function."""
from typing import Any, Callable

import numpy as np
from numpy.random import Generator

from moptipy.api.operators import Op0
from moptipy.utils.types import type_error


class Op0Forward(Op0):
    """Get the forwarder function."""

    def __init__(self):
        """Initialize this operator."""
        #: the internal blueprint for filling permutations
        self.__call: Callable[[Any], None] | None = None

    def op0(self, random: Generator, dest: np.ndarray) -> None:
        """Forward the call."""
        self.__call(dest)

    def forward_to(self, call: Callable[[Any], None]) -> None:
        """
        Set the `Callable` to forward all calls from :meth:`op0` to.

        :param call: the `Callable` to which all calls to :meth:`op0` should
            be delegated to.
        """
        if not callable(call):
            raise type_error(call, "call", call=True)
        self.__call = call

    def stop_forwarding(self) -> None:
        """Stop forwarding the call."""
        self.__call = None

    def __str__(self) -> str:
        """
        Get the name of this operator.

        :return: "forward"
        """
        return "forward"



"""
Provide the functionality to access search and solution spaces.

A :class:`Space` is the abstraction of the data structures for solutions and
points in the search space that need to be generated, copied, and stored
during the optimization process. This allows us to develop black-box
algorithms while still being able to properly remember the best solutions,
storing them as text strings in log files, and to validate whether they are
correct.

All search or solution spaces in `moptipy` inherit from :class:`Space`. If
you implement a new space, you should test it with the pre-defined unit test
routine :func:`~moptipy.tests.space.validate_space`.

The following pre-defined spaces are currently available:

- :class:`~moptipy.spaces.bitstrings.BitStrings`, the space of `n`-dimensional
  bit strings
- :class:`~moptipy.spaces.intspace.IntSpace`, a space of `n`-dimensional
  integer strings, where each element is between predefined inclusive bounds
  `min_value...max_value`.
- :class:`~moptipy.spaces.permutations.Permutations` is a special version of
  the :class:`~moptipy.spaces.intspace.IntSpace` where all elements are
  permutations of a base string
  :attr:`~moptipy.spaces.permutations.Permutations.blueprint`. This means that
  it can represent permutations both with and without repetitions. Depending
  on the base string, each element may occur an element-specific number of
  times. For the base string `(-1, -1, 2, 7, 7, 7)`, for example, `-1` may
  occur twice, `2` can occur once, and `7` three times.
- :class:`~moptipy.spaces.vectorspace.VectorSpace` is the space of
  `n`-dimensional floating point number vectors.
"""

from typing import Any

from moptipy.api.component import Component
from moptipy.utils.types import type_error


# start book
class Space(Component):
    """
    A class to represent both search and solution spaces.

    The space basically defines a container data structure and basic
    operations that we can apply to them. For example, a solution
    space contains all the possible solutions to an optimization
    problem. All of them are instances of one data structure. An
    optimization as well as a black-box process needs to be able to
    create and copy such objects. In order to store the solutions we
    found in a text file, we must further be able to translate them to
    strings. We should also be able to parse such strings. It is also
    important to detect whether two objects are the same and whether
    the contents of an object are valid. All of this functionality is
    offered by the `Space` class.
    """

    def create(self) -> Any:
        # end book
        """
        Generate an instance of the data structure managed by the space.

        The state/contents of this data structure are undefined. It may
        not pass the :meth:`validate` method.

        :return: the new instance
        """

    def copy(self, dest, source) -> None:  # +book
        """
        Copy one instance of the data structure to another one.

        :param dest: the destination data structure,
            whose contents will be overwritten with those from `source`
        :param source: the source data structure, which remains
            unchanged and whose contents will be copied to `dest`
        """

    def to_str(self, x) -> str:  # +book
        """
        Obtain a textual representation of an instance of the data structure.

        This method should convert an element of the space to a string
        representation that is parseable by :meth:from_str: and should ideally
        not be too verbose. For example, when converting a list or array `x`
        of integers to a string, one could simply do
        `";".join([str(xx) for xx in x])`, which would convert it to a
        semicolon-separated list without any wasted space.

        :param x: the instance
        :return: the string representation of x
        """

    def is_equal(self, x1, x2) -> bool:  # +book
        """
        Check if the contents of two instances of the data structure are equal.

        :param x1: the first instance
        :param x2: the second instance
        :return: `True` if the contents are equal, `False` otherwise
        """

    def from_str(self, text: str) -> Any:  # +book
        """
        Transform a string `text` to one element of the space.

        This method should be implemented as inverse to :meth:to_str:.
        It should check the validity of the result before returning it.
        It may not always be possible to implement this method, but you
        should try.

        :param text: the input string
        :return: the element in the space corresponding to `text`
        """

    def validate(self, x) -> None:  # +book
        """
        Check whether a given point in the space is valid.

        :param x: the point
        :raises ValueError: if the point `x` is invalid
        :raises TypeError: if the point `x` (or one of its elements, if
            applicable) has the wrong data type
        """

    def n_points(self) -> int:
        """
        Get the approximate number of different elements in the space.

        This operation can help us when doing tests of the space API
        implementations. If we know how many points exist in the space,
        then we can judge whether a method that randomly generates
        points is sufficiently random, for instance.

        By default, this method simply returns `2`. If you have a better
        approximation of the size of the space, then you should override it.

        :return: the approximate scale of the space
        """
        return 2


def check_space(space: Space | None,
                none_is_ok: bool = False) -> Space | None:
    """
    Check whether an object is a valid instance of :class:`Space`.

    :param space: the object
    :param none_is_ok: is it ok if `None` is passed in?
    :return: the object
    :raises TypeError: if `space` is not an instance of
        :class:`~moptipy.api.space.Space`
    """
    if not isinstance(space, Space):
        if none_is_ok and (space is None):
            return None
        raise type_error(space, "space", Space)
    return space



"""
The weighted sum scalarization of multi-objective problems.

Here we provide two simple methods to scalarize multi-objective problems by
using weights, namely

- :class:`~moptipy.mo.problem.weighted_sum.WeightedSum`, a sum with arbitrary,
  user-defined weights of the objective values
- :class:`~moptipy.mo.problem.weighted_sum.Prioritize`, a weighted sum of the
  objective values where the weights are automatically determined such that
  the first objective function is prioritized over the second one, the second
  one over the third, and so on.
"""
from math import inf, isfinite
from typing import Any, Callable, Final, Iterable, cast

import numpy as np
from numpy import sum as npsum

from moptipy.api.mo_utils import dominates
from moptipy.api.objective import Objective
from moptipy.mo.problem.basic_mo_problem import BasicMOProblem
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.math import try_int
from moptipy.utils.nputils import dtype_for_data
from moptipy.utils.strings import num_to_str
from moptipy.utils.types import type_error


def _sum_int(a: np.ndarray) -> int:
    """
    Sum up an array and convert the result to an `int` value.

    :param a: the array
    :returns: the sum of the elements in `a` as `int`
    """
    return int(npsum(a))


def _sum_float(a: np.ndarray) -> float:
    """
    Sum up an array and convert the result to a `float` value.

    :param a: the array
    :returns: the sum of the elements in `a` as `float`
    """
    return float(npsum(a))


class BasicWeightedSum(BasicMOProblem):
    """
    Base class for scalarizing objective values by a weighted sum.

    This class brings the basic tools to scalarize vectors of objective
    values by computing weighted sums. This class should not be used
    directly. Instead, use its sub-classes
    :class:`~moptipy.mo.problem.weighted_sum.WeightedSum` and
    :class:`~moptipy.mo.problem.weighted_sum.Prioritize`.
    """

    def __init__(self, objectives: Iterable[Objective],
                 get_scalarizer: Callable[
                     [bool, int, list[int | float],
                      list[int | float], Callable[
                          [None | np.dtype | tuple[int | float, ...]], None]],
                     Callable[[np.ndarray], int | float]],
                 domination: Callable[[np.ndarray, np.ndarray], int] | None
                 = dominates) -> None:
        """
        Create the sum-based scalarization.

        :param objectives: the objectives
        :param domination: a function reflecting the domination relationship
            between two vectors of objective values. It must obey the contract
            of :meth:`~moptipy.api.mo_problem.MOProblem.f_dominates`, which is
            the same as :func:`moptipy.api.mo_utils.dominates`, to which it
            defaults. `None` overrides nothing.
        """
        holder: list[Any] = []
        super().__init__(
            objectives,
            cast(Callable[[bool, int, list[int | float],
                           list[int | float]], Callable[
                [np.ndarray], int | float]],
                lambda ai, n, lb, ub, fwd=holder.append:
                get_scalarizer(ai, n, lb, ub, fwd)),
            domination)
        if len(holder) != 2:
            raise ValueError(
                f"need weights and weights dtype, but got {holder}.")
        #: the internal weights
        self.weights: Final[tuple[int | float, ...] | None] = \
            cast(tuple[int | float, ...] | None, holder[0])
        if self.weights is not None:
            if not isinstance(self.weights, tuple):
                raise type_error(self.weights, "weights", [tuple, None])
            if len(self.weights) != self.f_dimension():
                raise ValueError(
                    f"length of weights {self.weights} is not "
                    f"f_dimension={self.f_dimension()}.")
        #: the internal weights dtype
        self.__weights_dtype: Final[np.dtype | None] = \
            cast(np.dtype | None, holder[1])
        if self.__weights_dtype is not None:
            if not isinstance(self.__weights_dtype, np.dtype):
                raise type_error(
                    self.__weights_dtype, "weights_dtype", np.dtype)

    def __str__(self):
        """
        Get the string representation of the weighted sum scalarization.

        :returns: `"weightedSumBase"`
        """
        return "weightedSumBase"

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of this function to the provided destination.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)

        weights: tuple[int | float, ...] | None = self.weights
        logger.key_value("weights", ";".join(
            (["1"] * self.f_dimension()) if weights is None else
            [num_to_str(w) for w in weights]))
        logger.key_value("weightsDtype",
                         "None" if self.__weights_dtype is None
                         else self.__weights_dtype.char)


def _make_sum_scalarizer(
        always_int: bool, n: int,
        lower_bounds: list[int | float], upper_bounds: list[int | float],
        weights: tuple[int | float, ...] | None,
        callback: Callable[[None | np.dtype | tuple[
            int | float, ...]], None]) -> Callable[[np.ndarray], int | float]:
    """
    Create a weighted sum scalarization function.

    If `weights` is `None`, we will just use the plain summation function from
    numpy and convert its result to `int` if `always_int` is `True` and to
    `float` otherwise.
    If `weights` is a tuple of weights, then we will convert it to a numpy
    `ndarray`. If `always_int` is `True` and all weights are integers, and the
    lower and upper bound of the objectives are known, we try to pick the
    smallest integer data type for this array big enough to hold both the
    total lower and total upper bound. If at least the lower bounds are >= 0,
    then we pick the largest unsigned integer data type. If the bounds are
    unknown, then we pick the largest signed integer type. If any weight is
    not an integer and `always_int` is `False`, we use a default floating
    point weight array.

    This should yield the overall fastest, most precise, and most memory
    efficient way to compute a weighted sum scalarization.

    :param always_int: will all objectives always be integer
    :param n: the number of objectives
    :param lower_bounds: the optional lower bounds
    :param upper_bounds: the optional upper bounds
    :param weights: the optional array of weights, `None` if all weights
        are `1`.
    :param callback: the callback function to receive the weights
    :returns: the scalarization function
    """
    if not isinstance(lower_bounds, list):
        raise type_error(lower_bounds, "lower_bounds", list)
    if len(lower_bounds) != n:
        raise ValueError(
            f"there should be {n} values in lower_bounds={lower_bounds}")

    if not isinstance(upper_bounds, list):
        raise type_error(upper_bounds, "upper_bounds", list)
    if len(upper_bounds) != n:
        raise ValueError(
            f"there should be {n} values in upper_bounds={lower_bounds}")

    if weights is None:
        callback(None)
        callback(None)
        return _sum_int if always_int else _sum_float
    if not isinstance(weights, tuple):
        raise type_error(weights, "weights", tuple)
    if len(weights) != n:
        raise ValueError(
            f"there should be {n} values in weights={lower_bounds}")
    if not isinstance(always_int, bool):
        raise type_error(always_int, "always_int", bool)

    min_sum: int | float = 0
    max_sum: int | float = 0
    min_weight: int | float = inf
    max_weight: int | float = -inf
    everything_is_int: bool = always_int

    for i, weight in enumerate(weights):
        if weight <= 0:
            raise ValueError("no weight can be <=0, but encountered "
                             f"{weight} in {weights}.")

        if not isinstance(weight, int):
            everything_is_int = False
            if not isfinite(weight):
                raise ValueError("weight must be finite, but "
                                 f"encountered {weight} in {weights}.")
            min_sum = -inf
            max_sum = inf
            break

        if weight < min_weight:
            min_weight = weight
        if weight > max_weight:
            max_weight = weight

        if lower_bounds is not None:
            min_sum += weight * lower_bounds[i]
        if upper_bounds is not None:
            max_sum += weight * upper_bounds[i]

    if min_sum >= max_sum:
        raise ValueError(
            f"weighted sum minimum={min_sum} >= maximum={max_sum}?")

    # re-check for plain summation
    if 1 <= min_weight <= max_weight <= 1:
        callback(None)
        callback(None)
        return _sum_int if always_int else _sum_float

    dtype: Final[np.dtype] = dtype_for_data(
        everything_is_int, min_sum, max_sum)
    use_weights: Final[np.ndarray] = np.array(weights, dtype)

    callback(weights)
    callback(dtype)

    if everything_is_int:
        return cast(Callable[[np.ndarray], int | float],
                    lambda a, w=use_weights: int(npsum(a * w)))
    return cast(Callable[[np.ndarray], int | float],
                lambda a, w=use_weights: float(npsum(a * w)))


class WeightedSum(BasicWeightedSum):
    """Scalarize objective values by computing their weighted sum."""

    def __init__(self, objectives: Iterable[Objective],
                 weights: Iterable[int | float] | None = None,
                 domination: Callable[[np.ndarray, np.ndarray], int] | None
                 = dominates) -> None:
        """
        Create the sum-based scalarization.

        :param objectives: the objectives
        :param weights: the weights of the objective values, or `None` if all
            weights are `1`.
        :param domination: a function reflecting the domination relationship
            between two vectors of objective values. It must obey the contract
            of :meth:`~moptipy.api.mo_problem.MOProblem.f_dominates`, which is
            the same as :func:`moptipy.api.mo_utils.dominates`, to which it
            defaults. `None` overrides nothing.
        """
        use_weights: tuple[int | float, ...] | None \
            = None if weights is None else tuple(try_int(w) for w in weights)

        super().__init__(
            objectives,
            cast(Callable[
                 [bool, int, list[int | float],
                  list[int | float], Callable[
                      [None | np.dtype | tuple[int | float, ...]], None]],
                 Callable[[np.ndarray], int | float]],
                 lambda ai, n, lb, ub, cb, uw=use_weights:
                 _make_sum_scalarizer(ai, n, lb, ub, uw, cb)),
            domination)

    def __str__(self):
        """
        Get the string representation of the weighted sum scalarization.

        :returns: `"weightedSum"`
        """
        return "weightedSum" if self.f_dominates is dominates \
            else "weightedSumWithDominationFunc"


def _prioritize(
        always_int: bool, n: int,
        lower_bounds: list[int | float], upper_bounds: list[int | float],
        callback: Callable[[None | np.dtype | tuple[
            int | float, ...]], None]) -> Callable[[np.ndarray], int | float]:
    """
    Create a weighted-sum based prioritization of the objective functions.

    If all objective functions are integers and have upper and lower bounds,
    we can use integer weights to create a prioritization such that gaining
    one unit of the first objective function is always more important than any
    improvement of the second objective, that gaining one unit of the second
    objective always outweighs all possible gains in terms of the third one,
    and so on.

    :param always_int: will all objectives always be integer
    :param n: the number of objectives
    :param lower_bounds: the optional lower bounds
    :param upper_bounds: the optional upper bounds
    :param callback: the callback function to receive the weights
    :returns: the scalarization function
    """
    if n == 1:
        return _make_sum_scalarizer(always_int, n, lower_bounds,
                                    upper_bounds, None, callback)
    if not always_int:
        raise ValueError("priority-based weighting is only possible for "
                         "integer-valued objectives")
    weights: list[int] = [1]
    weight: int = 1
    for i in range(n - 1, 0, -1):
        lb: int | float = lower_bounds[i]
        if not isinstance(lb, (int, float)):
            raise type_error(lb, f"lower_bound[{i}]", (int, float))
        if not isfinite(lb):
            raise ValueError(f"lower_bound[{i}]={lb}, but must be finite")
        if not isinstance(lb, int):
            raise type_error(lb, f"finite lower_bound[{i}]", int)
        ub: int | float = upper_bounds[i]
        if not isinstance(ub, (int, float)):
            raise type_error(ub, f"upper_bound[{i}]", (int, float))
        if not isfinite(ub):
            raise ValueError(f"upper_bound[{i}]={ub}, but must be finite")
        if not isinstance(ub, int):
            raise type_error(ub, f"finite upper_bound[{i}]", int)
        weight *= (1 + cast(int, ub) - min(0, cast(int, lb)))
        weights.append(weight)

    weights.reverse()
    return _make_sum_scalarizer(always_int, n, lower_bounds, upper_bounds,
                                tuple(weights), callback)


class Prioritize(BasicWeightedSum):
    """Prioritize the first objective over the second and so on."""

    def __init__(self, objectives: Iterable[Objective],
                 domination: Callable[[np.ndarray, np.ndarray], int] | None
                 = dominates) -> None:
        """
        Create the sum-based prioritization.

        :param objectives: the objectives
        :param domination: a function reflecting the domination relationship
            between two vectors of objective values. It must obey the contract
            of :meth:`~moptipy.api.mo_problem.MOProblem.f_dominates`, which is
            the same as :func:`moptipy.api.mo_utils.dominates`, to which it
            defaults. `None` overrides nothing.
        """
        super().__init__(objectives, _prioritize, domination)

    def __str__(self):
        """
        Get the name of the weighted sum-based prioritization.

        :returns: `"weightBasedPrioritization"`
        """
        return "weightBasedPrioritization" if self.f_dominates is dominates \
            else "weightBasedPrioritizationWithDominationFunc"



"""
The FFA-based version of the (1+1)-EA: the (1+1)-FEA.

This algorithm is based on :class:`~moptipy.algorithms.so.rls.RLS`, i.e., the
(1+1)-EA, but uses Frequency Fitness Assignment (FFA) as fitness assignment
process. FFA replaces all objective values with their encounter frequencies in
the selection decisions. The more often an objective value is encountered, the
higher gets its encounter frequency. Therefore, local optima are slowly
receiving worse and worse fitness.

1. Thomas Weise, Zhize Wu, Xinlu Li, and Yan Chen. Frequency Fitness
   Assignment: Making Optimization Algorithms Invariant under Bijective
   Transformations of the Objective Function Value. *IEEE Transactions on
   Evolutionary Computation* 25(2):307-319. April 2021. Preprint available at
   arXiv:2001.01416v5 [cs.NE] 15 Oct 2020.
   https://dx.doi.org/10.1109/TEVC.2020.3032090
2. Thomas Weise, Zhize Wu, Xinlu Li, Yan Chen, and Jörg Lässig. Frequency
   Fitness Assignment: Optimization without Bias for Good Solutions can be
   Efficient. *IEEE Transactions on Evolutionary Computation (TEVC)*. 2022.
   Early Access. https://dx.doi.org/10.1109/TEVC.2022.3191698
3. Thomas Weise, Mingxu Wan, Ke Tang, Pu Wang, Alexandre Devert, and Xin
   Yao. Frequency Fitness Assignment. *IEEE Transactions on Evolutionary
   Computation (IEEE-EC),* 18(2):226-243, April 2014.
   https://dx.doi.org/10.1109/TEVC.2013.2251885
4. Thomas Weise, Yan Chen, Xinlu Li, and Zhize Wu. Selecting a diverse set of
   benchmark instances from a tunable model problem for black-box discrete
   optimization algorithms. *Applied Soft Computing Journal (ASOC),*
   92:106269, June 2020. https://dx.doi.org/10.1016/j.asoc.2020.106269
5. Thomas Weise, Xinlu Li, Yan Chen, and Zhize Wu. Solving Job Shop Scheduling
   Problems Without Using a Bias for Good Solutions. In *Genetic and
   Evolutionary Computation Conference Companion (GECCO'21 Companion),*
   July 10-14, 2021, Lille, France. ACM, New York, NY, USA.
   ISBN 978-1-4503-8351-6. https://dx.doi.org/10.1145/3449726.3463124
6. Thomas Weise, Mingxu Wan, Ke Tang, and Xin Yao. Evolving Exact Integer
   Algorithms with Genetic Programming. In *Proceedings of the IEEE Congress
   on Evolutionary Computation (CEC'14), Proceedings of the 2014 World
   Congress on Computational Intelligence (WCCI'14),* pages 1816-1823,
   July 6-11, 2014, Beijing, China. Los Alamitos, CA, USA: IEEE Computer
   Society Press. ISBN: 978-1-4799-1488-3.
   https://dx.doi.org/10.1109/CEC.2014.6900292
"""
from typing import Callable, Final, cast

import numpy as np
from numpy.random import Generator

from moptipy.api.algorithm import Algorithm1
from moptipy.api.operators import Op0, Op1
from moptipy.api.process import Process


class FEA1plus1(Algorithm1):
    """
    The FFA-based version of the (1+1)-EA: the (1+1)-FEA.

    This algorithm applies Frequency Fitness Assignment (FFA).
    This means that it does not select solutions based on whether
    they are better or worse. Instead, it selects the solution whose
    objective value has been encountered during the search less often.
    The word "best" therefore is not used in the traditional sense, i.e.,
    that one solution is better than another one terms of its objective
    value. Instead, the current best solution is always the one whose
    objective value we have seen the least often.

    In each step, a (1+1)-FEA creates a modified copy `new_x` of the
    current best solution `best_x`. It then increments the frequency fitness
    of both solutions by 1. If frequency fitness of `new_x` is not bigger
    the one of `best_x`, it becomes the new `best_x`.
    Otherwise, it is discarded.

    This algorithm implementation requires that objective values are
    integers and have lower and upper bounds that are not too far
    away from each other.
    """

    def __init__(self, op0: Op0, op1: Op1) -> None:
        """
        Create the (1+1)-FEA.

        :param op0: the nullary search operator
        :param op1: the unary search operator
        """
        super().__init__("fea1p1", op0, op1)

    def solve(self, process: Process) -> None:
        """
        Apply the (1+1)-FEA to an optimization problem.

        :param process: the black-box process object
        """
        # Create records for old and new point in the search space.
        best_x = process.create()  # record for best-so-far solution
        new_x = process.create()  # record for new solution
        lb: Final[int] = cast(int, process.lower_bound())

        # h holds the encounter frequency of each objective value.
        h: Final[np.ndarray] = np.zeros(
            cast(int, process.upper_bound()) - lb + 1, np.uint64)
        # Obtain the random number generator.
        random: Final[Generator] = process.get_random()

        # Put function references in variables to save time.
        evaluate: Final[Callable] = process.evaluate  # the objective
        op1: Final[Callable] = self.op1.op1  # the unary operator
        should_terminate: Final[Callable] = process.should_terminate

        # Start at a random point in the search space and evaluate it.
        self.op0.op0(random, best_x)  # Create 1 solution randomly and
        best_f: int = cast(int, evaluate(best_x)) - lb  # evaluate it.

        while not should_terminate():  # Until we need to quit...
            op1(random, new_x, best_x)  # new_x = neighbor of best_x
            new_f: int = cast(int, evaluate(new_x)) - lb

            h[new_f] = h[new_f] + 1  # Increase frequency of new_f and
            h[best_f] = best_h = h[best_f] + 1  # of best_f.
            if h[new_f] <= best_h:  # new_x is no worse than best_x?
                best_f = new_f  # Store its objective value.
                best_x, new_x = new_x, best_x  # Swap best and new.



"""
Selection algorithms are common modules that choose `n` out of `N` objects.

Selection algorithms are modules of the fully-configurable Evolutionary
Algorithm :class:`~moptipy.algorithms.so.general_ea.GeneralEA`. They can
utilize fitness values computed by the fitness assignment processes
(:class:`~moptipy.algorithms.so.fitness.Fitness`). Of course, they can
also be applied in different contexts and are not bound to single-objective
optimization.

:class:`~moptipy.algorithms.modules.selection.Selection` is especially
important in Evolutionary Algorithms
(`~moptipy.algorithms.so.general_ea.GeneralEA`), where it is used in two
places: As *survival selection*, it chooses which points will be allowed to
remain in the population and, hence, survive into the mating pool for the next
generation. As *mating selection* methods, they choose the inputs of the
search operations from the mating pool.

:class:`~moptipy.algorithms.modules.selection.Selection` algorithms must
*only* use the
:attr:`~moptipy.algorithms.modules.selection.FitnessRecord.fitness` of a
solution record (and random numbers) to make their decisions. These fitness
values are subject to minimization. They can equal to the objective values in
optimization or stem from a :class:`~moptipy.algorithms.so.fitness.Fitness`
Assignment Process.

The following selection algorithms have currently been implemented:

- :class:`~moptipy.algorithms.modules.selections.best.Best` selection
  selects the best `n` solutions without replacement. This is a common
  strategy for survival selection, especially in (mu+lambda) EAs
  (compatible to :class:`~moptipy.algorithms.so.ea.EA`).
- :class:`~moptipy.algorithms.modules.selections.tournament.Tournament`
  selection conducts a tournament with `k` contestants for of the `n` slots
  in the destination and the winners of the tournaments are chosen.
- :class:`~moptipy.algorithms.modules.selections.random_without_replacement\
.RandomWithoutReplacement` selects random solutions without replacement. It is
  a common strategy for mating selection.
- :class:`~moptipy.algorithms.modules.selections.fitness_proportionate_sus\
.FitnessProportionateSUS` performs fitness proportionate selection for
  minimization using stochastic uniform sampling and, optionally, a minimum
  selection probability threshold. It is the classic survival selection
  algorithm in Genetic Algorithm.
"""
from typing import Any, Callable, Protocol

from numpy.random import Generator

from moptipy.api.component import Component
from moptipy.utils.types import type_error


# start book
class FitnessRecord(Protocol):
    """A fitness record stores data together with a fitness."""

# end book
    #: the fitness value, the only criterion to be used by a selection
    #: algorithm (besides random numbers)
    fitness: int | float

    def __lt__(self, other) -> bool:
        """
        Compare the fitness of this record with the fitness of another one.

        :param other: the other fitness record
        """
        return self.fitness < other.fitness
# start book


class Selection(Component):
    """The base class for selections algorithms."""

    def select(self, source: list[FitnessRecord],
               dest: Callable[[FitnessRecord], Any],
               n: int, random: Generator) -> None:  # pylint: disable=W0613
        """
        Select `n` records from `source` and pass them to `dest`.

        When choosing the `n` records from `source` to be passed to
        `dest`, only the :attr:`~FitnessRecord.fitness` attribute of
        the records and the random numbers from `random` must be used
        as decision criteria.

        :param source: the list with the records to select from
        :param dest: the destination collector Callable to invoke for
            each selected record, can be :class:`list`.`append`.
        :param n: the number of records to select
        :param random: the random number generator
        """
# end book


def check_selection(selection: Selection) -> Selection:
    """
    Check whether an object is a valid instance of :class:`Selection`.

    :param selection: the Selection object
    :return: the object
    :raises TypeError: if `selections` is not an instance of
        :class:`Selection` or if it is an instance of the abstract base
        class
    """
    if not isinstance(selection, Selection):
        raise type_error(selection, "selections", Selection)
    if selection.__class__ is Selection:
        raise TypeError("cannot use abstract class 'Selection' directly")
    return selection



"""
An operator swapping several elements in a permutation.

This operator works similarly to
:class:`~moptipy.operators.permutations.op1_swap2.Op1Swap2`, but instead of
swapping two elements (i.e., performing 1 swap), it will perform a random
number of swaps.

First, the operator copies the source string `x` into the destination `dest`.
It then chooses a random index `i1`. It remembers the element at that index
in two variables, `last` and `first`. (`last` will always be the value of the
element will next be overwritten and `first` will never change and remember
the first overwritten element, i.e., the value that we need to store again
into the array in the end.) We also initialize a variable `continue_after`
with `True`. This variable will always tell us if we need to continue and
draw another random index.

We then begin with the main loop, which is iterated as long as
`continue_after` is `True`. Directly as first action in this loop, we set
`continue_after = ri(2) <= 0`. `ri(2)` will return a random integer in
`{0, 1}`. `ri(2) <= 0` will therefore be `True` with probability 0.5.

We now will draw a new random index `i2`. We load the element at index `i2`
into variable `current`. If `current == last`, we immediately go back and
draw another random `i2`. The reason is that we want to store `current` at
`i1`, i.e., where `last` is currently stored. If `current == last`, this
would change nothing. Furthermore, if `continue_after` is `False`, then
`current` must also be different from `first`: If we exit the main loop,
then we will store `first` into the place where we found `current`. Remember,
we will overwrite the element at index `i1` with `current`, so the very first
element we overwrite must eventually be stored back into the array.

OK, if we have obtained an index `i2` whose corresponding element `current`
fulfills all requirements, we can set `dest[i1] = last = current`, i.e.,
remember `current` in `last` and also store it at index `i1`. Next, we will
overwrite the element at index `i2`, so we set `i1 = i2`. If `continue_after`
is `True`, the loop will continue. Otherwise, it will stop.

In the latter case, we store `first` back into the array at the last index
`i1` and do `dest[i1] = first`.

As a result, the operator will swap exactly 2 elements with probability 0.5.
With probability 0.25, it will swap three elements, with 0.125 probability, it
will swap 4 elements, and so on.

1. Thomas Weise. *Optimization Algorithms.* 2021. Hefei, Anhui, China:
   Institute of Applied Optimization (IAO), School of Artificial Intelligence
   and Big Data, Hefei University. http://thomasweise.github.io/oa/
"""
from typing import Callable, Final

import numpy as np
from numpy.random import Generator

from moptipy.api.operators import Op1


# start book
class Op1SwapN(Op1):
    """
    This unary search operation swaps several (different) elements.

    It is similar to `swap2`, but instead may perform a random number
    of swaps. After each swap, it decides with probability 0.5 whether
    or not to perform another swap.
    """

    def op1(self, random: Generator,
            dest: np.ndarray, x: np.ndarray) -> None:
        """
        Copy `x` into `dest` and then swap several different values.

        :param random: the random number generator
        :param dest: the array to receive the modified copy of `x`
        :param x: the existing point in the search space
        """
        np.copyto(dest, x)  # First, we copy `x` to `dest`.
        length: Final[int] = len(dest)  # Get the length of `dest`.
        ri: Final[Callable[[int], int]] = random.integers  # fast call!

        i1: int = ri(length)  # Get the first random index.
        last = first = dest[i1]  # Get the value at the first index.
        continue_after: bool = True  # True -> loop at least once.
        while continue_after:  # Repeat until we should stop
            continue_after = ri(2) <= 0  # Continue after iteration?
            while True:  # Loop forever until eligible element found.
                i2: int = ri(length)  # Get a new random index.
                current = dest[i2]  # Get the value at the new index.
                if current == last:  # If it is the same as the
                    continue  # previous value, continue.
                if continue_after or (current != first):  # If we want
                    break  # to stop, then it must be != first value.
            dest[i1] = last = current  # Store value for from i2 at i1.
            i1 = i2  # Update i1 to now point to cell of i2.
        dest[i1] = first  # Finally, store first element back at end.
    # end book

    def __str__(self) -> str:
        """
        Get the name of this unary operator.

        :returns: "swapn", the name of this operator
        :retval "swapn": always
        """
        return "swapn"



"""
The Job Shop Scheduling Problem is a good example for optimization tasks.

The JSSP is one of the most well-known combinatorial optimization tasks.
Here we provide a set of benchmark instances and examples for solving this
problem.
"""


def __lang_setup() -> None:
    """Perform the language setup."""
    from moptipy.utils.lang import DE, EN, ZH  # pylint: disable=C0415

    # the English language strings
    EN.extend({
        "gantt_info": "{gantt.instance.name} ({gantt.instance.jobs} "
                      "jobs \u00D7 {gantt.instance.machines} machines), "
                      "makespan {gantt[:,:,2].max()}",
        "gantt_info_no_ms": "{gantt.instance.name} ({gantt.instance.jobs} "
                            "jobs \u00D7 {gantt.instance.machines} machines)",
        "gantt_info_short": "{gantt.instance.name} / {gantt[:,:,2].max()}",
        "machine": "machine",
        "makespan": "makespan"
    })
    # the German language strings
    DE.extend({
        "gantt_info": "{gantt.instance.name} ({gantt.instance.jobs} "
                      "Jobs \u00D7 {gantt.instance.machines} Maschinen), "
                      "Makespan {gantt[:,:,2].max()}",
        "gantt_info_no_ms": "{gantt.instance.name} ({gantt.instance.jobs} "
                            "Jobs \u00D7 {gantt.instance.machines} Maschinen)",
        "gantt_info_short": "{gantt.instance.name} / {gantt[:,:,2].max()}",
        "machine": "Maschine",
        "makespan": "Makespan"
    })
    # the Chinese language strings
    ZH.extend({
        "gantt_info": "{gantt.instance.name} ({gantt.instance.jobs}份作业"
                      "\u00D7{gantt.instance.machines}台机器), "
                      "最大完工时间{gantt[:,:,2].max()}",
        "gantt_info_no_ms": "{gantt.instance.name} ({gantt.instance.jobs}份作业"
                            "\u00D7{gantt.instance.machines}台机器) ",
        "gantt_info_short": "{gantt.instance.name} / {gantt[:,:,2].max()}",
        "machine": "机器",
        "makespan": "最大完工时间"
    })


__lang_setup()  # invoke the language setup
del __lang_setup  # delete the language setup routine



"""A fitness representing the rank of a solution based on its quality."""

from math import inf

from numpy.random import Generator

from moptipy.algorithms.so.fitness import Fitness, FRecord


# start book
class Rank(Fitness):
    """A fitness computing the rank of an individual based on its quality."""

    def assign_fitness(self, p: list[FRecord], random: Generator) -> None:
        """
        Assign the rank fitness.

        :param p: the list of records
        :param random: ignored

        >>> l = [FRecord(0, 10), FRecord(1, 5), FRecord(2, 5), FRecord(3, -1)]
        >>> from numpy.random import default_rng
        >>> Rank().assign_fitness(l, default_rng())
        >>> l[0].x
        3
        >>> l[0].fitness
        1
        >>> l[1].x
        1
        >>> l[1].fitness
        2
        >>> l[2].x
        2
        >>> l[2].fitness
        2
        >>> l[3].x
        0
        >>> l[3].fitness
        4
        """
        for rec in p:  # first copy rec.f to rec.fitness
            rec.fitness = rec.f  # because then we can easily sort
        p.sort()  # sort based on objective values

        rank: int = -1  # the rank counter
        last_fitness: int | float = -inf
        for i, rec in enumerate(p):
            v = rec.fitness
            if v > last_fitness:  # if fitness differs, step rank
                rank = i + 1  # +1 so smallest-possible rank is 1
                last_fitness = v
            rec.fitness = rank  # assign the rank
# end book

    def __str__(self):
        """
        Get the name of this default fitness assignment.

        :return: the name of this fitness assignment strategy
        :retval "rank": always
        """
        return "rank"



"""Plot a set of `Progress` or `StatRun` objects into one figure."""
from math import isfinite
from typing import Any, Callable, Final, Iterable

from matplotlib.artist import Artist  # type: ignore
from matplotlib.axes import Axes  # type: ignore
from matplotlib.figure import Figure, SubplotBase  # type: ignore

import moptipy.utils.plot_defaults as pd
import moptipy.utils.plot_utils as pu
from moptipy.evaluation.axis_ranger import AxisRanger
from moptipy.evaluation.base import get_algorithm, get_instance, sort_key
from moptipy.evaluation.progress import Progress
from moptipy.evaluation.stat_run import StatRun, get_statistic
from moptipy.evaluation.styler import Styler
from moptipy.utils.lang import Lang
from moptipy.utils.types import type_error


def plot_progress(
        progresses: Iterable[Progress | StatRun],
        figure: SubplotBase | Figure,
        x_axis: AxisRanger | Callable[[str], AxisRanger] =
        AxisRanger.for_axis,
        y_axis: AxisRanger | Callable[[str], AxisRanger] =
        AxisRanger.for_axis,
        legend: bool = True,
        distinct_colors_func: Callable[[int], Any] = pd.distinct_colors,
        distinct_line_dashes_func: Callable[[int], Any] =
        pd.distinct_line_dashes,
        importance_to_line_width_func: Callable[[int], float] =
        pd.importance_to_line_width,
        importance_to_alpha_func: Callable[[int], float] =
        pd.importance_to_alpha,
        importance_to_font_size_func: Callable[[int], float] =
        pd.importance_to_font_size,
        x_grid: bool = True,
        y_grid: bool = True,
        x_label: None | str | Callable[[str], str] = Lang.translate,
        x_label_inside: bool = True,
        x_label_location: float = 0.5,
        y_label: None | str | Callable[[str], str] = Lang.translate,
        y_label_inside: bool = True,
        y_label_location: float = 1.0,
        inst_priority: float = 0.666,
        algo_priority: float = 0.333,
        stat_priority: float = 0.0,
        instance_sort_key: Callable[[str], Any] = lambda x: x,
        algorithm_sort_key: Callable[[str], Any] = lambda x: x,
        stat_sort_key: Callable[[str], Any] = lambda x: x,
        color_algorithms_as_fallback_group: bool = True,
        instance_namer: Callable[[str], str] = lambda x: x,
        algorithm_namer: Callable[[str], str] = lambda x: x) -> Axes:
    """
    Plot a set of progress or statistical run lines into one chart.

    :param progresses: the iterable of progresses and statistical runs
    :param figure: the figure to plot in
    :param x_axis: the x_axis ranger
    :param y_axis: the y_axis ranger
    :param legend: should we plot the legend?
    :param distinct_colors_func: the function returning the palette
    :param distinct_line_dashes_func: the function returning the line styles
    :param importance_to_line_width_func: the function converting importance
        values to line widths
    :param importance_to_alpha_func: the function converting importance
        values to alphas
    :param importance_to_font_size_func: the function converting importance
        values to font sizes
    :param x_grid: should we have a grid along the x-axis?
    :param y_grid: should we have a grid along the y-axis?
    :param x_label: a callable returning the label for the x-axis, a label
        string, or `None` if no label should be put
    :param x_label_inside: put the x-axis label inside the plot (so that
        it does not consume additional vertical space)
    :param x_label_location: the location of the x-axis label
    :param y_label: a callable returning the label for the y-axis, a label
        string, or `None` if no label should be put
    :param y_label_inside: put the y-axis label inside the plot (so that
        it does not consume additional horizontal space)
    :param y_label_location: the location of the y-axis label
    :param inst_priority: the style priority for instances
    :param algo_priority: the style priority for algorithms
    :param stat_priority: the style priority for statistics
    :param instance_sort_key: the sort key function for instances
    :param algorithm_sort_key: the sort key function for algorithms
    :param stat_sort_key: the sort key function for statistics
    :param color_algorithms_as_fallback_group: if only a single group of data
        was found, use algorithms as group and put them in the legend
    :param instance_namer: the name function for instances receives an
        instance ID and returns an instance name; default=identity function
    :param algorithm_namer: the name function for algorithms receives an
        algorithm ID and returns an algorithm name; default=identity function
    :returns: the axes object to allow you to add further plot elements
    """
    # Before doing anything, let's do some type checking on the parameters.
    # I want to ensure that this function is called correctly before we begin
    # to actually process the data. It is better to fail early than to deliver
    # some incorrect results.
    if not isinstance(progresses, Iterable):
        raise type_error(progresses, "progresses", Iterable)
    if not isinstance(figure, (SubplotBase, Figure)):
        raise type_error(figure, "figure", (SubplotBase, Figure))
    if not isinstance(legend, bool):
        raise type_error(legend, "legend", bool)
    if not callable(distinct_colors_func):
        raise type_error(
            distinct_colors_func, "distinct_colors_func", call=True)
    if not callable(distinct_colors_func):
        raise type_error(
            distinct_colors_func, "distinct_colors_func", call=True)
    if not callable(distinct_line_dashes_func):
        raise type_error(
            distinct_line_dashes_func, "distinct_line_dashes_func", call=True)
    if not callable(importance_to_line_width_func):
        raise type_error(importance_to_line_width_func,
                         "importance_to_line_width_func", call=True)
    if not callable(importance_to_alpha_func):
        raise type_error(
            importance_to_alpha_func, "importance_to_alpha_func", call=True)
    if not callable(importance_to_font_size_func):
        raise type_error(importance_to_font_size_func,
                         "importance_to_font_size_func", call=True)
    if not isinstance(x_grid, bool):
        raise type_error(x_grid, "x_grid", bool)
    if not isinstance(y_grid, bool):
        raise type_error(y_grid, "y_grid", bool)
    if not ((x_label is None) or callable(x_label)
            or isinstance(x_label, str)):
        raise type_error(x_label, "x_label", (str, None), call=True)
    if not isinstance(x_label_inside, bool):
        raise type_error(x_label_inside, "x_label_inside", bool)
    if not isinstance(x_label_location, float):
        raise type_error(x_label_location, "x_label_location", float)
    if not ((y_label is None) or callable(y_label)
            or isinstance(y_label, str)):
        raise type_error(y_label, "y_label", (str, None), call=True)
    if not isinstance(y_label_inside, bool):
        raise type_error(y_label_inside, "y_label_inside", bool)
    if not isinstance(y_label_location, float):
        raise type_error(y_label_location, "y_label_location", float)
    if not isinstance(inst_priority, float):
        raise type_error(inst_priority, "inst_priority", float)
    if not isfinite(inst_priority):
        raise ValueError(f"inst_priority cannot be {inst_priority}.")
    if not isinstance(algo_priority, float):
        raise type_error(algo_priority, "algo_priority", float)
    if not isfinite(algo_priority):
        raise ValueError(f"algo_priority cannot be {algo_priority}.")
    if not isinstance(stat_priority, float):
        raise type_error(stat_priority, "stat_priority", float)
    if not isfinite(stat_priority):
        raise ValueError(f"stat_priority cannot be {stat_priority}.")
    if not callable(instance_sort_key):
        raise type_error(instance_sort_key, "instance_sort_key", call=True)
    if not callable(algorithm_sort_key):
        raise type_error(algorithm_sort_key, "algorithm_sort_key", call=True)
    if not callable(stat_sort_key):
        raise type_error(stat_sort_key, "stat_sort_key", call=True)
    if not callable(instance_namer):
        raise type_error(instance_namer, "instance_namer", call=True)
    if not callable(algorithm_namer):
        raise type_error(algorithm_namer, "algorithm_namer", call=True)
    if not isinstance(color_algorithms_as_fallback_group, bool):
        raise type_error(color_algorithms_as_fallback_group,
                         "color_algorithms_as_fallback_group", bool)

    # First, we try to find groups of data to plot together in the same
    # color/style. We distinguish progress objects from statistical runs.
    instances: Final[Styler] = Styler(key_func=get_instance,
                                      namer=instance_namer,
                                      none_name=Lang.translate("all_insts"),
                                      priority=inst_priority,
                                      name_sort_function=instance_sort_key)
    algorithms: Final[Styler] = Styler(key_func=get_algorithm,
                                       namer=algorithm_namer,
                                       none_name=Lang.translate("all_algos"),
                                       priority=algo_priority,
                                       name_sort_function=algorithm_sort_key)
    statistics: Final[Styler] = Styler(key_func=get_statistic,
                                       none_name=Lang.translate("single_run"),
                                       priority=stat_priority,
                                       name_sort_function=stat_sort_key)
    x_dim: str | None = None
    y_dim: str | None = None
    progress_list: list[Progress] = []
    statrun_list: list[StatRun] = []

    # First pass: find out the statistics, instances, algorithms, and types
    for prg in progresses:
        instances.add(prg)
        algorithms.add(prg)
        statistics.add(prg)
        if isinstance(prg, Progress):
            progress_list.append(prg)
        elif isinstance(prg, StatRun):
            statrun_list.append(prg)
        else:
            raise type_error(prg, "progress plot element",
                             (Progress, StatRun))

        # Validate that we have consistent time and objective units.
        if x_dim is None:
            x_dim = prg.time_unit
        elif x_dim != prg.time_unit:
            raise ValueError(
                f"Time units {x_dim} and {prg.time_unit} do not fit!")

        if y_dim is None:
            y_dim = prg.f_name
        elif y_dim != prg.f_name:
            raise ValueError(
                f"F-units {y_dim} and {prg.f_name} do not fit!")
    del progresses

    if (len(progress_list) + len(statrun_list)) <= 0:
        raise ValueError("Empty input data?")

    if (x_dim is None) or (y_dim is None):
        raise ValueError("Illegal state?")

    instances.finalize()
    algorithms.finalize()
    statistics.finalize()

    # pick the right sorting order
    sf: Callable[[StatRun | Progress], Any] = sort_key
    if (instances.count > 1) and (algorithms.count == 1) \
            and (statistics.count == 1):
        def __x(r: StatRun | Progress, ssf=instance_sort_key) -> str:
            return ssf(r.instance)
        sf = __x
    elif (instances.count == 1) and (algorithms.count > 1) \
            and (statistics.count == 1):
        def __x(r: StatRun | Progress, ssf=algorithm_sort_key) -> str:
            return ssf(r.instance)
        sf = __x
    elif (instances.count == 1) and (algorithms.count == 1) \
            and (statistics.count > 1):
        def __x(r: StatRun | Progress, ssf=stat_sort_key) -> str:
            return ssf(r.instance)
        sf = __x

    statrun_list.sort(key=sf)
    progress_list.sort()

    def __set_importance(st: Styler) -> None:
        if st is statistics:
            none = -1
            not_none = 1
        else:
            none = 1
            not_none = 0
        none_lw = importance_to_line_width_func(none)
        not_none_lw = importance_to_line_width_func(not_none)
        st.set_line_width(lambda x: [none_lw if i <= 0 else not_none_lw
                                     for i in range(x)])
        none_a = importance_to_alpha_func(none)
        not_none_a = importance_to_alpha_func(not_none)
        st.set_line_alpha(lambda x: [none_a if i <= 0 else not_none_a
                                     for i in range(x)])

    # determine the style groups
    groups: list[Styler] = []

    no_importance = True
    if instances.count > 1:
        groups.append(instances)
    if algorithms.count > 1:
        groups.append(algorithms)
    add_stat_to_groups = False
    if statistics.count > 1:
        if statistics.has_none and (statistics.count == 2):
            __set_importance(statistics)
            no_importance = False
            add_stat_to_groups = True
        else:
            groups.append(statistics)

    if len(groups) > 0:
        groups.sort()
        groups[0].set_line_color(distinct_colors_func)

        if len(groups) > 1:
            groups[1].set_line_dash(distinct_line_dashes_func)

            if (len(groups) > 2) and no_importance:
                g = groups[2]
                if g.count > 2:
                    raise ValueError(
                        f"Cannot have {g.count} importance values.")
                __set_importance(g)
                no_importance = False
    elif color_algorithms_as_fallback_group:
        algorithms.set_line_color(distinct_colors_func)
        groups.append(algorithms)

    if add_stat_to_groups:
        groups.append(statistics)

    # If we only have <= 2 groups, we can mark None and not-None values with
    # different importance.
    if no_importance and statistics.has_none and (statistics.count > 1):
        __set_importance(statistics)
        no_importance = False
    if no_importance and instances.has_none and (instances.count > 1):
        __set_importance(instances)
        no_importance = False
    if no_importance and algorithms.has_none and (algorithms.count > 1):
        __set_importance(algorithms)

    # we will collect all lines to plot in plot_list
    plot_list: list[dict] = []

    # first we collect all progress object
    for prgs in progress_list:
        style = pd.create_line_style()
        for g in groups:
            g.add_line_style(prgs, style)
        style["x"] = prgs.time
        style["y"] = prgs.f
        plot_list.append(style)
    del progress_list

    # now collect the plot data for the statistics
    for sn in statistics.keys:
        if sn is None:
            continue
        for sr in statrun_list:
            if statistics.key_func(sr) != sn:
                continue

            style = pd.create_line_style()
            for g in groups:
                g.add_line_style(sr, style)
            style["x"] = sr.stat[:, 0]
            style["y"] = sr.stat[:, 1]
            plot_list.append(style)
    del statrun_list

    font_size_0: Final[float] = importance_to_font_size_func(0)

    # set up the graphics area
    axes: Final[Axes] = pu.get_axes(figure)
    axes.tick_params(axis="x", labelsize=font_size_0)
    axes.tick_params(axis="y", labelsize=font_size_0)

    # draw the grid
    if x_grid or y_grid:
        grid_lwd = importance_to_line_width_func(-1)
        if x_grid:
            axes.grid(axis="x", color=pd.GRID_COLOR, linewidth=grid_lwd)
        if y_grid:
            axes.grid(axis="y", color=pd.GRID_COLOR, linewidth=grid_lwd)

    # set up the axis rangers
    if callable(x_axis):
        x_axis = x_axis(x_dim)
    if not isinstance(x_axis, AxisRanger):
        raise type_error(x_axis, "x_axis", AxisRanger)

    if callable(y_axis):
        y_axis = y_axis(y_dim)
    if not isinstance(y_axis, AxisRanger):
        raise type_error(y_axis, "y_axis", AxisRanger)

    # plot the lines
    for line in plot_list:
        axes.step(where="post", **line)
        x_axis.register_array(line["x"])
        y_axis.register_array(line["y"])
    del plot_list

    x_axis.apply(axes, "x")
    y_axis.apply(axes, "y")

    if legend:
        handles: list[Artist] = []

        for g in groups:
            g.add_to_legend(handles.append)
            g.has_style = False

        if instances.has_style:
            instances.add_to_legend(handles.append)
        if algorithms.has_style:
            algorithms.add_to_legend(handles.append)
        if statistics.has_style:
            statistics.add_to_legend(handles.append)

        if len(handles) > 0:
            axes.legend(loc="upper right",
                        handles=handles,
                        labelcolor=[art.color if hasattr(art, "color")
                                    else pd.COLOR_BLACK for art in handles],
                        fontsize=font_size_0)

    pu.label_axes(axes=axes,
                  x_label=x_label(x_dim) if callable(x_label) else x_label,
                  x_label_inside=x_label_inside,
                  x_label_location=x_label_location,
                  y_label=y_label(y_dim) if callable(y_label) else y_label,
                  y_label_inside=y_label_inside,
                  y_label_location=y_label_location,
                  font_size=font_size_0)
    return axes



"""An objective function counting the leading ones in a bit string."""
from typing import Final

import numba  # type: ignore
import numpy as np

from moptipy.api.objective import Objective
from moptipy.utils.types import type_error


@numba.njit(nogil=True, cache=True)
def leadingones(x: np.ndarray) -> int:
    """
    Get the length of the string minus the number of leading ones.

    :param x: the np array
    :return: the number of leading ones

    >>> print(leadingones(np.array([False, False, True, False, False])))
    5
    >>> print(leadingones(np.array([True, False, False, True, True])))
    4
    >>> print(leadingones(np.array([True, True, False, False, False])))
    3
    >>> print(leadingones(np.array([True, True, True, False, True])))
    2
    >>> print(leadingones(np.array([True, True, True, True, False])))
    1
    >>> print(leadingones(np.array([True, True, True, True, True])))
    0
    """
    xlen: Final[int] = len(x)
    for i in range(xlen):
        if not x[i]:
            return xlen - i
    return 0


class LeadingOnes(Objective):
    """Maximize the number of leadings ones in a bit string."""

    def __init__(self, n: int) -> None:  # +book
        """
        Initialize the leading ones objective function.

        :param n: the dimension of the problem

        >>> print(LeadingOnes(55).n)
        55
        >>> print(LeadingOnes(4).evaluate(np.array([True, True, True, False])))
        1
        """
        super().__init__()
        if not isinstance(n, int):
            raise type_error(n, "n", int)
        #: the upper bound = the length of the bit strings
        self.n: Final[int] = n
        self.evaluate = leadingones  # type: ignore

    def lower_bound(self) -> int:
        """
        Get the lower bound of the leadingones objective function.

        :return: 0

        >>> print(LeadingOnes(3).lower_bound())
        0
        """
        return 0

    def upper_bound(self) -> int:
        """
        Get the upper bound of the leadingones objective function.

        :return: the length of the bit string

        >>> print(LeadingOnes(5).upper_bound())
        5
        """
        return self.n

    def is_always_integer(self) -> bool:
        """
        Return `True` because :func:`leadingones` always returns `int` values.

        :retval True: always
        """
        return True

    def __str__(self) -> str:
        """
        Get the name of the leadingones objective function.

        :return: `leadingones_` + lenth of string

        >>> print(LeadingOnes(10))
        leadingones_10
        """
        return f"leadingones_{self.n}"



"""The one-dimensional Ising problem."""
from typing import Final

import numba  # type: ignore
import numpy as np

from moptipy.api.objective import Objective
from moptipy.utils.types import type_error


@numba.njit(nogil=True, cache=True)
def ising1d(x: np.ndarray) -> int:
    """
    Compute the objective value of the 1-dimensional Ising problem.

    :param x: the np array
    :return: the trap function value

    >>> print(ising1d(np.array([True, True, True, True, True])))
    0
    >>> print(ising1d(np.array([False, False, False, False, False])))
    0
    >>> print(ising1d(np.array([False, False, False, True, False])))
    2
    >>> print(ising1d(np.array([True, False, False, False, False])))
    2
    >>> print(ising1d(np.array([False, False, False, False, True])))
    2
    >>> print(ising1d(np.array([True, False, False, False, True])))
    2
    >>> print(ising1d(np.array([True, False, True, False, False])))
    4
    >>> print(ising1d(np.array([True, False, True, False, True, False])))
    6
    """
    s: int = 0
    prev: bool = x[-1]
    for cur in x:
        if cur != prev:
            s += 1
        prev = cur
    return s


class Ising1d(Objective):
    """The one-dimensional Ising problem."""

    def __init__(self, n: int) -> None:  # +book
        """
        Initialize the one-dimensional Ising problem.

        :param n: the dimension of the problem

        >>> print(Ising1d(7).n)
        7
        >>> print(Ising1d(3).evaluate(np.array([True, False, True])))
        2
        """
        super().__init__()
        if not isinstance(n, int):
            raise type_error(n, "n", int)
        #: the upper bound = the length of the bit strings
        self.n: Final[int] = n
        self.evaluate = ising1d  # type: ignore

    def lower_bound(self) -> int:
        """
        Get the lower bound of the one-dimensional Ising problem.

        :return: 0

        >>> print(Ising1d(7).lower_bound())
        0
        """
        return 0

    def upper_bound(self) -> int:
        """
        Get the upper bound of the one-dimensional Ising problem.

        :return: the length of the bit string

        >>> print(Ising1d(12).upper_bound())
        12
        """
        return self.n

    def is_always_integer(self) -> bool:
        """
        Return `True` because :func:`ising1d` always returns `int` values.

        :retval True: always
        """
        return True

    def __str__(self) -> str:
        """
        Get the name of the one-dimensional Ising problem.

        :return: `ising1d_` + length of string

        >>> print(Ising1d(5))
        ising1d_5
        """
        return f"ising1d_{self.n}"



"""Generate random mock experiment parameters."""

from dataclasses import dataclass
from math import ceil, isfinite
from typing import Any, Final, Iterable

from numpy.random import Generator

from moptipy.utils.console import logger
from moptipy.utils.nputils import rand_generator, rand_seeds_from_str
from moptipy.utils.strings import sanitize_name
from moptipy.utils.types import type_error


def fixed_random_generator() -> Generator:
    """
    Get the single, fixed random generator for the dummy experiment API.

    :returns: the random number generator
    """
    if not hasattr(fixed_random_generator, "gen"):
        setattr(fixed_random_generator, "gen", rand_generator(1))
    return getattr(fixed_random_generator, "gen")


def _random_name(namelen: int,
                 random: Generator = fixed_random_generator()) -> str:
    """
    Generate a random name of a given length.

    :param namelen: the length of the name
    :param random: a random number generator
    :returns: a name of the length
    """
    if not isinstance(namelen, int):
        raise type_error(namelen, "namelen", int)
    if namelen <= 0:
        raise ValueError(f"namelen must be > 0, but is {namelen}.")
    namer: Final[tuple[str, str, str]] = ("bcdfghjklmnpqrstvwxyz", "aeiou",
                                          "0123456789")
    name = ["x"] * namelen
    index: int = 0
    n_done: bool = False
    for i in range(namelen):
        namee = namer[index]
        name[i] = namee[random.integers(len(namee))]
        n_done = n_done or (i == 2)
        if index <= 0:
            index += 1
        else:
            if index == 2:
                if random.integers(2) <= 0:
                    continue
            index = int((index + 1 + random.integers(2)) % len(namer))
            if n_done:
                while index == 2:
                    index = 1 - min(1, random.integers(6))

    return "".join(name)


def __append_not_allowed(forbidden,
                         dest: set[str | float | int]) -> None:
    """
    Append items to the set of not-allowed values.

    :param forbidden: the forbidden elements
    :param dest: the set to append to
    """
    if not forbidden:
        return
    if isinstance(forbidden, (str, int, float)):
        dest.add(forbidden)
    elif isinstance(forbidden, Instance):
        dest.add(forbidden.name)
        dest.add(forbidden.hardness)
        dest.add(forbidden.jitter)
        dest.add(forbidden.scale)
        dest.add(forbidden.best)
        dest.add(forbidden.worst)
    elif isinstance(forbidden, Algorithm):
        dest.add(forbidden.name)
        dest.add(forbidden.strength)
        dest.add(forbidden.jitter)
    elif isinstance(forbidden, Iterable):
        for item in forbidden:
            __append_not_allowed(item, dest)
    else:
        raise type_error(forbidden, "element to add",
                         (str, int, float, Instance, Algorithm, Iterable))


def _make_not_allowed(forbidden: Iterable[str | float | int | Iterable[
        Any]] | None = None) -> set[str | float | int]:
    """
    Create a set of not-allowed values.

    :param forbidden: the forbidden elements
    :returns: the set of not-allowed values
    """
    not_allowed: set[Any] = set()
    __append_not_allowed(forbidden, not_allowed)
    return not_allowed


@dataclass(frozen=True, init=False, order=True)
class Instance:
    """An immutable instance description record."""

    #: The instance name.
    name: str
    #: The instance hardness, in (0, 1), larger values are worst
    hardness: float
    #: The instance jitter, in (0, 1), larger values are worst
    jitter: float
    #: The instance scale, in (0, 1), larger values are worst
    scale: float
    #: The best (smallest) possible objective value
    best: int
    #: The worst (largest) possible objective value
    worst: int
    #: The set of attractors, i.e., local optima - including best and worst
    attractors: tuple[int, ...]

    def __init__(self,
                 name: str,
                 hardness: float,
                 jitter: float,
                 scale: float,
                 best: int,
                 worst: int,
                 attractors: tuple[int, ...]):
        """
        Create a mock problem instance description.

        :param name: the instance name
        :param hardness: the instance hardness
        :param jitter: the instance jitter
        :param scale: the instance scale
        :param best: the best (smallest) possible objective value
        :param worst: the worst (largest) possible objective value
        :param attractors: the set of attractors, i.e., local
            optima - including best and worst
        """
        if not isinstance(name, str):
            raise type_error(name, "name", str)
        if name != sanitize_name(name):
            raise ValueError(f"Invalid name '{name}'.")
        object.__setattr__(self, "name", name)

        if not isinstance(hardness, float):
            raise type_error(hardness, "hardness", float,)
        if (not isfinite(hardness)) or (hardness <= 0) or (hardness >= 1):
            raise ValueError(
                f"hardness must be in (0, 1), but is {hardness}.")
        object.__setattr__(self, "hardness", hardness)

        if not isinstance(jitter, float):
            raise type_error(jitter, "jitter", float)
        if (not isfinite(jitter)) or (jitter <= 0) or (jitter >= 1):
            raise ValueError(
                f"jitter must be in (0, 1), but is {jitter}.")
        object.__setattr__(self, "jitter", jitter)

        if not isinstance(scale, float):
            raise type_error(scale, "scale", float)
        if (not isfinite(scale)) or (scale <= 0) or (scale >= 1):
            raise ValueError(
                f"scale must be in (0, 1), but is {scale}.")
        object.__setattr__(self, "scale", scale)

        if not isinstance(best, int):
            raise type_error(best, "best", int)
        if (best <= 0) or (best >= 1_000_000_000):
            raise ValueError(
                f"best must be in 1...999999999, but is {best}.")
        object.__setattr__(self, "best", best)

        if not isinstance(worst, int):
            raise type_error(worst, "worst", int)
        if (worst <= (best + 7)) or (worst >= 1_000_000_000):
            raise ValueError(
                f"worst must be in {best + 8}...999999999, but is {worst}.")
        object.__setattr__(self, "worst", worst)

        if not isinstance(attractors, tuple):
            raise type_error(attractors, "attractors", tuple)
        if len(attractors) < 4:
            raise ValueError("attractors must contain at least 2 values,"
                             f" but contains only {len(attractors)}.")
        if attractors[0] != best:
            raise ValueError(
                f"attractors[0] must be {best}, but is {attractors[0]}")
        if attractors[-1] != worst:
            raise ValueError(
                f"attractors[-1] must be {worst}, but is {attractors[-1]}")
        prev = -1
        for att in attractors:
            if not isinstance(att, int):
                raise type_error(att, "each attractor", int)
            if (att < best) or (att > worst) or (att <= prev):
                raise ValueError(f"{att} not permitted after {prev} "
                                 f"for best={best} and worst={worst}.")
            prev = att
        object.__setattr__(self, "attractors", attractors)

    @staticmethod
    def create(n: int,
               forbidden: Any | None = None,
               random: Generator = fixed_random_generator()) \
            -> tuple["Instance", ...]:
        """
        Create a set of fixed problem instances.

        :param n: the number of instances to generate
        :param random: a random number generator
        :param forbidden: the forbidden names and hardnesses
        :returns: a tuple of instances
        """
        if not isinstance(n, int):
            raise type_error(n, "n", int)
        if n <= 0:
            raise ValueError(f"n must be > 0, but is {n}.")
        logger(f"now creating {n} instances.")

        not_allowed: Final[set[str | float]] = \
            _make_not_allowed(forbidden)
        names: list[str] = []
        hardnesses: list[float] = []
        jitters: list[float] = []
        scales: list[float] = []

        # First we choose a unique name.
        max_name_len: int = int(max(2, ceil(n / 6)))
        trials: int = 0
        while len(names) < n:
            trials += 1
            if trials > 1000:
                trials = 1
                max_name_len += 1
            nv = _random_name(int(3 + random.integers(max_name_len)), random)
            if nv not in not_allowed:
                names.append(nv)
                not_allowed.add(nv)

        # Now we pick an instance hardness.
        limit: int = 2000
        trials = 0
        while len(hardnesses) < n:
            v: float = -1
            while (v <= 0) or (v >= 1) or (not isfinite(v)):
                trials += 1
                if trials > 1000:
                    trials = 0
                    limit *= 2
                if random.integers(4) <= 0:
                    v = int(random.uniform(1, limit)) / limit
                else:
                    v = int(random.normal(loc=0.5, scale=0.25) * limit) / limit
            if v not in not_allowed:
                hardnesses.append(v)
                not_allowed.add(v)

        # Now we pick an instance jitter.
        limit = 2000
        trials = 0
        while len(jitters) < n:
            v = -1
            while (v <= 0) or (v >= 1) or (not isfinite(v)):
                trials += 1
                if trials > 1000:
                    trials = 0
                    limit *= 2
                if random.integers(4) <= 0:
                    v = int(random.uniform(1, limit)) / limit
                else:
                    v = int(random.normal(loc=0.5, scale=0.2) * limit) / limit
            if v not in not_allowed:
                jitters.append(v)
                not_allowed.add(v)

        # Now we choose a scale.
        limit = 2000
        trials = 0
        while len(scales) < n:
            v = -1
            while (v <= 0) or (v >= 1) or (not isfinite(v)):
                trials += 1
                if trials > 1000:
                    trials = 0
                    limit *= 2
                if random.integers(4) <= 0:
                    v = int(random.uniform(1, limit)) / limit
                else:
                    v = int(random.normal(loc=0.5, scale=0.2) * limit) / limit
            if v not in not_allowed:
                scales.append(v)
                not_allowed.add(v)

        # We choose the global optimum and the worst objective value.
        trials = 0
        scale: int = 1000
        loc: int = 1000
        limits: list[int] = []
        while len(limits) < (2 * n):
            tbound: int = -1
            while (tbound <= 0) or (tbound >= 1_000_000_000):
                trials += 1
                if trials > 1000:
                    trials = 0
                    scale += (scale // 3)
                    loc *= 2
                tbound = int(random.normal(
                    loc=random.integers(low=1, high=4) * loc, scale=scale))
            permitted: bool = True
            for b in limits:
                if abs(b - tbound) <= 41:
                    permitted = False
                    break
            if permitted and (tbound not in not_allowed):
                limits.append(tbound)
                not_allowed.add(tbound)

        result: list[Instance] = []
        attdone: set[int] = set()

        for i in range(n, 0, -1):
            trials = 0
            b1 = b2 = -1
            while trials < 10:
                trials += 1
                b1 = limits.pop(random.integers(i * 2))
                b2 = limits.pop(random.integers(i * 2 - 1))
                if b1 > b2:
                    b1, b2 = b2, b1
                if i <= 1:
                    break
                if (b1 * max(2.0, 0.4 * (10 - trials))) < b2:
                    break
                limits.append(b1)
                limits.append(b2)
            attdone.clear()
            attdone.add(b1)
            attdone.add(b2)

            # Now we make sure that there are at least 6 attractors.
            trials = 0
            min_dist = max(7.0, 0.07 * (b2 - b1))
            while ((len(attdone) < 6) or (random.integers(7) > 0)) \
                    and (trials < 20000):
                a = -1
                while ((a <= b1) or (a >= b2) or (a in attdone)) \
                        and (trials < 20000):
                    trials += 1
                    a = int(random.integers(low=b1 + 1, high=b2 - 1))
                if (a <= b1) or (a >= b2):
                    continue
                ok = True
                for aa in attdone:
                    if abs(aa - a) < min_dist:
                        ok = False
                        break
                if ok:
                    attdone.add(a)
                else:
                    if (trials % 1000) <= 0:
                        min_dist = max(5, 0.5 * min_dist)

            result.append(Instance(
                name=names.pop(random.integers(i)),
                hardness=hardnesses.pop(random.integers(i)),
                jitter=jitters.pop(random.integers(i)),
                scale=scales.pop(random.integers(i)),
                best=b1,
                worst=b2,
                attractors=tuple(sorted(attdone))))
        result.sort()
        logger(f"finished creating {n} instances.")
        return tuple(result)


@dataclass(frozen=True, init=False, order=True)
class Algorithm:
    """An immutable algorithm description record."""

    #: The algorithm name.
    name: str
    #: The algorithm strength, in (0, 1), larger values are worst
    strength: float
    #: The algorithm jitter, in (0, 1), larger values are worst
    jitter: float
    #: The algorithm complexity, in (0, 1), larger values are worst
    complexity: float

    def __init__(self,
                 name: str,
                 strength: float,
                 jitter: float,
                 complexity: float):
        """
        Create a mock algorithm description record.

        :param name: the algorithm name
        :param strength: the algorithm strength
        :param jitter: the algorithm jitter
        :param complexity: the algorithm complexity
        """
        if not isinstance(name, str):
            raise type_error(name, "name", str)
        if name != sanitize_name(name):
            raise ValueError(f"Invalid name '{name}'.")
        object.__setattr__(self, "name", name)

        if not isinstance(strength, float):
            raise type_error(strength, "strength", float)
        if (not isfinite(strength)) or (strength <= 0) or (strength >= 1):
            raise ValueError(
                f"strength must be in (0, 1), but is {strength}.")
        object.__setattr__(self, "strength", strength)

        if not isinstance(jitter, float):
            raise type_error(jitter, "jitter", float)
        if (not isfinite(jitter)) or (jitter <= 0) or (jitter >= 1):
            raise ValueError(
                f"jitter must be in (0, 1), but is {jitter}.")
        object.__setattr__(self, "jitter", jitter)

        if not isinstance(complexity, float):
            raise type_error(complexity, "complexity", float)
        if (not isfinite(complexity)) or (complexity <= 0) \
                or (complexity >= 1):
            raise ValueError(
                f"complexity must be in (0, 1), but is {complexity}.")
        object.__setattr__(self, "complexity", complexity)

    @staticmethod
    def create(n: int,
               forbidden: Any | None = None,
               random: Generator = fixed_random_generator()) \
            -> tuple["Algorithm", ...]:
        """
        Create a set of fixed mock algorithms.

        :param n: the number of algorithms to generate
        :param random: a random number generator
        :param forbidden: the forbidden names and strengths and so on
        :returns: a tuple of algorithms
        """
        if not isinstance(n, int):
            raise type_error(n, "n", int)
        if n <= 0:
            raise ValueError(f"n must be > 0, but is {n}.")
        logger(f"now creating {n} algorithms.")

        not_allowed: Final[set[str | float]] = \
            _make_not_allowed(forbidden)
        names: list[str] = []
        strengths: list[float] = []
        jitters: list[float] = []
        complexities: list[float] = []

        prefixes: Final[tuple[str, ...]] = ("aco", "bobyqa", "cma-es", "de",
                                            "ea", "eda", "ga", "gp", "hc",
                                            "ma", "pso", "rs", "rw", "sa",
                                            "umda")
        suffixes: Final[tuple[str, ...]] = ("1swap", "2swap", "µ")

        max_name_len: int = int(max(2, ceil(n / 6)))
        trials: int = 0
        while len(names) < n:
            trials += 1
            if trials > 1000:
                trials = 1
                max_name_len += 1

            name_mode = random.integers(5)
            if name_mode < 2:
                nva = _random_name(int(3 + random.integers(max_name_len)),
                                   random)
                if nva in not_allowed:
                    continue
                if name_mode == 1:
                    nvb = suffixes[random.integers(len(suffixes))]
                    nv = f"{nva}_{nvb}"
                else:
                    nv = nva
                if nv in not_allowed:
                    continue
                not_allowed.add(nva)
                not_allowed.add(nv)
                names.append(nv)
                continue

            nva = prefixes[random.integers(len(prefixes))]
            if name_mode == 3:
                nvb = _random_name(int(3 + random.integers(
                    max_name_len)), random)
                if nvb in not_allowed:
                    continue
                nv = f"{nva}_{nvb}"
            elif name_mode == 4:
                nvb = suffixes[random.integers(len(suffixes))]
                nv = f"{nva}_{nvb}"
            else:
                nv = nva
                nvb = ""

            if nv in not_allowed:
                continue
            names.append(nv)
            not_allowed.add(nv)
            if name_mode == 3:
                not_allowed.add(nvb)

        limit: int = 2000
        trials = 0
        while len(strengths) < n:
            v: float = -1
            while (v <= 0) or (v >= 1) or (not isfinite(v)):
                trials += 1
                if trials > 1000:
                    trials = 0
                    limit *= 2
                if random.integers(4) <= 0:
                    v = int(random.uniform(1, limit)) / limit
                else:
                    v = int(random.normal(loc=0.5,
                                          scale=0.25) * limit) / limit
            if v not in not_allowed:
                strengths.append(v)
                not_allowed.add(v)

        limit = 2000
        trials = 0
        while len(jitters) < n:
            v = -1
            while (v <= 0) or (v >= 1) or (not isfinite(v)):
                trials += 1
                if trials > 1000:
                    trials = 0
                    limit *= 2
                if random.integers(4) <= 0:
                    v = int(random.uniform(1, limit)) / limit
                else:
                    v = int(random.normal(loc=0.5,
                                          scale=0.2) * limit) / limit
            if v not in not_allowed:
                jitters.append(v)
                not_allowed.add(v)

        limit = 2000
        trials = 0
        while len(complexities) < n:
            v = -1
            while (v <= 0) or (v >= 1) or (not isfinite(v)):
                trials += 1
                if trials > 1000:
                    trials = 0
                    limit *= 2
                if random.integers(4) <= 0:
                    v = int(random.uniform(1, limit)) / limit
                else:
                    v = int(random.normal(loc=0.5,
                                          scale=0.2) * limit) / limit
            if v not in not_allowed:
                complexities.append(v)
                not_allowed.add(v)

        result: list[Algorithm] = []
        for i in range(n, 0, -1):
            result.append(Algorithm(
                name=names.pop(random.integers(i)),
                strength=strengths.pop(random.integers(i)),
                jitter=jitters.pop(random.integers(i)),
                complexity=complexities.pop(random.integers(i))))
        result.sort()

        logger(f"finished creating {n} algorithms.")
        return tuple(result)


@dataclass(frozen=True, init=False, order=True)
class BasePerformance:
    """An algorithm applied to a problem instance description record."""

    #: The algorithm.
    algorithm: Algorithm
    #: The problem instance
    instance: Instance
    #: The base performance, in (0, 1), larger values are worst
    performance: float
    #: The performance jitter, in (0, 1), larger values are worst
    jitter: float
    #: The time per FE, in (0, 1), larger values are worst
    speed: float

    def __init__(self,
                 algorithm: Algorithm,
                 instance: Instance,
                 performance: float,
                 jitter: float,
                 speed: float):
        """
        Create a mock algorithm-instance application description record.

        :param algorithm: the algorithm
        :param instance: the instance
        :param performance: the base performance
        :param jitter: the performance jitter
        :param speed: the time required per FE
        """
        if not isinstance(algorithm, Algorithm):
            raise type_error(algorithm, "algorithm", Algorithm)
        object.__setattr__(self, "algorithm", algorithm)
        if not isinstance(instance, Instance):
            raise type_error(instance, "instance", Instance)
        object.__setattr__(self, "instance", instance)

        if not isinstance(performance, float):
            raise type_error(performance, "performance", float)
        if (not isfinite(performance)) or (performance <= 0) \
                or (performance >= 1):
            raise ValueError(
                f"performance must be in (0, 1), but is {performance}.")
        object.__setattr__(self, "performance", performance)

        if not isinstance(jitter, float):
            raise type_error(jitter, "jitter", float)
        if (not isfinite(jitter)) or (jitter <= 0) or (jitter >= 1):
            raise ValueError(
                f"jitter must be in (0, 1), but is {jitter}.")
        object.__setattr__(self, "jitter", jitter)

        if not isinstance(speed, float):
            raise type_error(speed, "speed", float)
        if (not isfinite(speed)) or (speed <= 0) or (speed >= 1):
            raise ValueError(
                f"speed must be in (0, 1), but is {speed}.")
        object.__setattr__(self, "speed", speed)

    @staticmethod
    def create(instance: Instance,
               algorithm: Algorithm,
               random: Generator = fixed_random_generator()) \
            -> "BasePerformance":
        """
        Compute the basic performance of an algorithm on a problem instance.

        :param instance: the instance tuple
        :param algorithm: the algorithm tuple
        :param random: the random number generator
        :returns: a tuple of the performance in (0, 1); bigger values are
            worse, and a jitter in (0, 1), where bigger values are worse
        """
        if not isinstance(instance, Instance):
            raise type_error(instance, "instance", Instance)
        if not isinstance(algorithm, Algorithm):
            raise type_error(algorithm, "algorithm", Algorithm)
        logger("now creating base performance for algorithm "
               f"{algorithm.name} on instance {instance.name}.")

        perf: float = -1
        granularity: Final[int] = 2000
        while (perf <= 0) or (perf >= 1):
            if random.integers(20) <= 0:
                perf = random.uniform(low=0, high=1)
            else:
                perf = random.normal(
                    loc=0.5 * (instance.hardness + algorithm.strength),
                    scale=0.2 * (instance.jitter + algorithm.jitter))
            perf = int(perf * granularity) / granularity

        jit: float = -1
        while (jit <= 0) or (jit >= 1):
            if random.integers(15) <= 0:
                jit = random.uniform(low=0, high=1)
            else:
                jit = random.normal(
                    loc=0.5 * (instance.jitter + algorithm.jitter),
                    scale=0.2 * (instance.jitter + algorithm.jitter))
            jit = int(jit * granularity) / granularity

        speed: float = -1
        while (speed <= 0) or (speed >= 1):
            if random.integers(20) <= 0:
                speed = random.uniform(low=0, high=1)
            else:
                speed = random.normal(
                    loc=0.5 * (instance.scale + algorithm.complexity),
                    scale=0.2 * (instance.scale + algorithm.complexity))
            speed = int(speed * granularity) / granularity

        bp: Final[BasePerformance] = BasePerformance(algorithm=algorithm,
                                                     instance=instance,
                                                     performance=perf,
                                                     jitter=jit,
                                                     speed=speed)
        logger("finished base performance "
               f"{bp.algorithm.name}@{bp.instance.name}.")
        return bp


def get_run_seeds(instance: Instance, n_runs: int) -> tuple[int, ...]:
    """
    Get the seeds for the runs.

    :param instance: the mock instance
    :param n_runs: the number of runs
    :returns: a tuple of seeds
    """
    if not isinstance(instance, Instance):
        raise type_error(instance, "instance", Instance)
    if not isinstance(n_runs, int):
        raise type_error(n_runs, "n_runs", int)
    if n_runs <= 0:
        raise ValueError(f"n_runs must be > 0, but is {n_runs}.")
    res: Final[tuple[int, ...]] = tuple(sorted(rand_seeds_from_str(
        string=instance.name, n_seeds=n_runs)))
    logger(f"finished creating {n_runs} seeds for instance {instance.name}.")
    return res


@dataclass(frozen=True, init=False, order=True)
class Experiment:
    """An immutable experiment description."""

    #: The instances.
    instances: tuple[Instance, ...]
    #: The algorithms.
    algorithms: tuple[Algorithm, ...]
    #: The applications of the algorithms to the instances.
    applications: tuple[BasePerformance, ...]
    #: The random seeds per instance.
    per_instance_seeds: tuple[tuple[int, ...]]
    #: the seeds per instance
    __seeds_per_inst: dict[str | Instance, tuple[int, ...]]
    #: the performance per algorithm
    __perf_per_algo: dict[str | Algorithm, tuple[BasePerformance, ...]]
    #: the performance per instance
    __perf_per_inst: dict[str | Instance, tuple[BasePerformance, ...]]
    #: the algorithm by names
    __algo_by_name: dict[str, Algorithm]
    #: the algorithm names
    algorithm_names: tuple[str, ...]
    #: the instance by names
    __inst_by_name: dict[str, Instance]
    #: the instance names
    instance_names: tuple[str, ...]

    def __init__(self,
                 instances: tuple[Instance, ...],
                 algorithms: tuple[Algorithm, ...],
                 applications: tuple[BasePerformance, ...],
                 per_instance_seeds: tuple[tuple[int, ...], ...]):
        """
        Create a mock experiment definition.

        :param algorithms: the algorithms
        :param instances: the instances
        :param applications: the applications of the algorithms to the
            instances
        :param per_instance_seeds: the seeds
        """
        if not isinstance(instances, tuple):
            raise type_error(instances, "instances", tuple)
        if len(instances) <= 0:
            raise ValueError("instances must not be empty.")
        inst_bn: dict[str, Instance] = {}
        for a in instances:
            if not isinstance(a, Instance):
                raise type_error(a, "element of instances", Instance)
            if a.name in inst_bn:
                raise ValueError(f"double instance name {a.name}.")
            inst_bn[a.name] = a
        object.__setattr__(self, "instances", instances)
        object.__setattr__(self, "_Experiment__inst_by_name", inst_bn)
        object.__setattr__(self, "instance_names",
                           tuple(sorted(inst_bn.keys())))

        if not isinstance(algorithms, tuple):
            raise type_error(algorithms, "algorithms", tuple)
        if len(algorithms) <= 0:
            raise ValueError("algorithms must not be empty.")
        algo_bn: dict[str, Algorithm] = {}
        for b in algorithms:
            if not isinstance(b, Algorithm):
                raise type_error(b, "element of algorithms", Algorithm)
            if b.name in algo_bn:
                raise ValueError(f"double algorithm name {b.name}.")
            if b.name in inst_bn:
                raise ValueError(f"instance/algorithm name {b.name} clash.")
            algo_bn[b.name] = b
        object.__setattr__(self, "algorithms", algorithms)
        object.__setattr__(self, "_Experiment__algo_by_name", algo_bn)
        object.__setattr__(self, "algorithm_names",
                           tuple(sorted(algo_bn.keys())))

        if not isinstance(applications, tuple):
            raise type_error(applications, "applications", tuple)
        if len(applications) != len(algorithms) * len(instances):
            raise ValueError(
                f"There must be {len(algorithms) * len(instances)} "
                f"applications, but found {len(applications)}.")

        perf_per_inst: dict[Instance, list[BasePerformance]] = {}
        perf_per_algo: dict[Algorithm, list[BasePerformance]] = {}

        done: set[str] = set()
        for c in applications:
            if not isinstance(c, BasePerformance):
                raise type_error(c, "element of applications", BasePerformance)
            s = c.algorithm.name + "+" + c.instance.name
            if s in done:
                raise ValueError(f"Encountered application {s} twice.")
            done.add(s)
            if c.algorithm in perf_per_algo:
                perf_per_algo[c.algorithm].append(c)
            else:
                perf_per_algo[c.algorithm] = [c]
            if c.instance in perf_per_inst:
                perf_per_inst[c.instance].append(c)
            else:
                perf_per_inst[c.instance] = [c]
        object.__setattr__(self, "applications", applications)

        pa: dict[str | Algorithm, tuple[BasePerformance, ...]] = {}
        for ax in algorithms:
            lax: list[BasePerformance] = perf_per_algo[ax]
            lax.sort()
            pa[ax.name] = pa[ax] = tuple(lax)
        pi: dict[str | Instance, tuple[BasePerformance, ...]] = {}
        for ix in instances:
            lix: list[BasePerformance] = perf_per_inst[ix]
            lix.sort()
            pi[ix.name] = pi[ix] = tuple(lix)
        object.__setattr__(self, "_Experiment__perf_per_algo", pa)
        object.__setattr__(self, "_Experiment__perf_per_inst", pi)

        if not isinstance(per_instance_seeds, tuple):
            raise type_error(per_instance_seeds, "per_instance_seeds", tuple)
        if len(per_instance_seeds) != len(instances):
            raise ValueError(
                f"There must be one entry for each of the {len(instances)} "
                "instances, but per_instance_seeds only "
                f"has {len(per_instance_seeds)}.")
        xl: int = -1
        inst_seeds: Final[dict[str | Instance, tuple[int, ...]]] = {}
        for idx, d in enumerate(per_instance_seeds):
            if not isinstance(d, tuple):
                raise type_error(d, "element of per_instance_seeds", tuple)
            if len(d) <= 0:
                raise ValueError(f"there must be at least one per "
                                 f"instance seed, but found {len(d)}.")
            if xl < 0:
                xl = len(d)
            if len(d) != xl:
                raise ValueError(f"there must be {xl} per "
                                 f"instance seeds, but found {len(d)}.")
            for e in d:
                if not isinstance(e, int):
                    raise type_error(e, "element of seeds", int)
            inst_seeds[instances[idx]] = d
            inst_seeds[instances[idx].name] = d
        object.__setattr__(self, "per_instance_seeds", per_instance_seeds)
        object.__setattr__(self, "_Experiment__seeds_per_inst", inst_seeds)

    @staticmethod
    def create(n_instances: int,
               n_algorithms: int,
               n_runs: int,
               random: Generator = fixed_random_generator()) -> "Experiment":
        """
        Create an experiment definition.

        :param n_instances: the number of instances
        :param n_algorithms: the number of algorithms
        :param n_runs: the number of per-instance runs
        :param random: the random number generator to use
        """
        if not isinstance(n_instances, int):
            raise type_error(n_instances, "n_instances", int)
        if n_instances <= 0:
            raise ValueError(
                f"n_instances must be > 0, but is {n_instances}.")
        if not isinstance(n_algorithms, int):
            raise type_error(n_algorithms, "n_algorithms", int)
        if n_algorithms <= 0:
            raise ValueError(
                f"n_algorithms must be > 0, but is {n_algorithms}.")
        if not isinstance(n_runs, int):
            raise type_error(n_runs, "n_runs", int)
        if n_algorithms <= 0:
            raise ValueError(
                f"n_runs must be > 0, but is {n_runs}.")
        logger(f"now creating mock experiment with {n_algorithms} algorithms "
               f"on {n_instances} instances for {n_runs} runs.")

        insts = Instance.create(n_instances, random=random)
        algos = Algorithm.create(n_algorithms, forbidden=insts, random=random)
        app = [BasePerformance.create(i, a, random)
               for i in insts for a in algos]
        app.sort()
        seeds = [get_run_seeds(i, n_runs) for i in insts]
        res: Final[Experiment] = Experiment(instances=insts,
                                            algorithms=algos,
                                            applications=tuple(app),
                                            per_instance_seeds=tuple(seeds))
        logger(f"finished creating mock experiment with {len(res.instances)} "
               f"instances, {len(res.algorithms)} algorithms, and "
               f"{len(res.per_instance_seeds[0])} runs per instance-"
               f"algorithm combination.")
        return res

    def seeds_for_instance(self, instance: str | Instance) \
            -> tuple[int, ...]:
        """
        Get the seeds for the specified instance.

        :param instance: the instance
        :returns: the seeds
        """
        return self.__seeds_per_inst[instance]

    def instance_applications(self, instance: str | Instance) \
            -> tuple[BasePerformance, ...]:
        """
        Get the applications of the algorithms to a specific instance.

        :param instance: the instance
        :returns: the applications
        """
        return self.__perf_per_inst[instance]

    def algorithm_applications(self, algorithm: str | Algorithm) \
            -> tuple[BasePerformance, ...]:
        """
        Get the applications of an algorithm to the instances.

        :param algorithm: the algorithm
        :returns: the applications
        """
        return self.__perf_per_algo[algorithm]

    def get_algorithm(self, name: str) -> Algorithm:
        """
        Get an algorithm by name.

        :param name: the algorithm name
        :returns: the algorithm instance
        """
        return self.__algo_by_name[name]

    def get_instance(self, name: str) -> Instance:
        """
        Get an instance by name.

        :param name: the instance name
        :returns: the instance
        """
        return self.__inst_by_name[name]



"""The latex text format driver."""

from io import TextIOBase
from typing import Final

from moptipy.utils.formatted_string import (
    NAN,
    NEGATIVE_INFINITY,
    NUMBER,
    POSITIVE_INFINITY,
    SPECIAL,
    TEXT,
)
from moptipy.utils.text_format import TextFormatDriver

#: the exponent prefix
_EPREFIX = r"\hspace*{0.15em}*\hspace*{0.1em}10\textsuperscript{"

#: special characters in LaTeX
SPECIAL_CHARS: Final[dict[str, str]] = {
    "\u2205": r"$\emptyset$",
    "\u221E": r"$\infty$",
    "-\u221E": r"$-\infty$",
    "inf": r"$\infty$",
    "-inf": r"$-\infty$",
    "nan": r"$\emptyset$",
    "\u03b1": r"$\alpha$",
    "\u2014": "---"
}


class LaTeX(TextFormatDriver):
    r"""
    The LaTeX text driver.

    >>> from io import StringIO
    >>> from moptipy.utils.formatted_string import FormattedStr
    >>> from moptipy.utils.table import Table
    >>> s = StringIO()
    >>> latex = LaTeX.instance()
    >>> print(str(latex))
    tex
    >>> with Table(s, "lrc", latex) as t:
    ...     with t.header() as header:
    ...         with header.row() as h:
    ...             h.cell(FormattedStr("1", bold=True))
    ...             h.cell(FormattedStr("2", code=True))
    ...             h.cell(FormattedStr("3", italic=True))
    ...     with t.section() as g:
    ...         with g.row() as r:
    ...             r.cell("a")
    ...             r.cell("b")
    ...             r.cell("c")
    ...         with g.row() as r:
    ...             r.cell("d")
    ...             r.cell("e")
    ...             r.cell("f")
    >>> print(f"'{s.getvalue()}'")
    '\begin{tabular}{lrc}%
    \hline%
    {\textbf{1}}&{\texttt{2}}&{\textit{3}}\\%
    \hline%
    a&b&c\\%
    d&e&f\\%
    \hline%
    \end{tabular}%
    '
    """

    def begin_table_body(self, stream: TextIOBase, cols: str) -> None:
        """Write the beginning of the table body."""
        stream.write(f"\\begin{{tabular}}{{{cols}}}%\n")

    def end_table_body(self, stream: TextIOBase, cols: str) -> None:
        """Write the ending of the table body."""
        stream.write("\\end{tabular}%\n")

    def begin_table_header(self, stream: TextIOBase, cols: str) -> None:
        """Begin the header of a LaTeX table."""
        stream.write("\\hline%\n")

    def end_table_header(self, stream: TextIOBase, cols: str) -> None:
        """End the header of a LaTeX table."""
        stream.write("\\hline%\n")

    def end_table_section(self, stream: TextIOBase, cols: str,
                          section_index: int, n_rows: int) -> None:
        """End a table section."""
        stream.write("\\hline%\n")

    def end_table_section_header(self, stream: TextIOBase, cols: str,
                                 section_index: int) -> None:
        """End a table section header."""
        stream.write("\\hline%\n")

    def end_table_row(self, stream: TextIOBase, cols: str,
                      section_index: int, row_index: int) -> None:
        """End a row in a LaTeX table."""
        stream.write("\\\\%\n")

    def begin_table_cell(self, stream: TextIOBase, cols: str,
                         section_index: int, row_index: int,
                         col_index: int, cell_mode: int) -> None:
        """Begin a LaTeX table cell."""
        if col_index > 0:
            stream.write("&")

    def text(self, stream: TextIOBase, text: str, bold: bool, italic: bool,
             code: bool, mode: int) -> None:
        """Print a text string."""
        if len(text) <= 0:
            return
        if bold:
            stream.write("{\\textbf{")
        if italic:
            stream.write("{\\textit{")
        if code:
            stream.write("{\\texttt{")

        if mode == TEXT:
            stream.write(text.replace("_", "\\_"))
        elif mode == NUMBER:
            i: int = text.find("e")
            if i < 0:
                i = text.find("E")
            if i > 0:
                stream.write(f"{text[:i]}{_EPREFIX}{text[i + 1:]}}}")
            else:
                stream.write(text.replace("_", "\\_"))
        elif mode == NAN:
            stream.write(r"$\emptyset$")
        elif mode == POSITIVE_INFINITY:
            stream.write(r"$\infty$")
        elif mode == NEGATIVE_INFINITY:
            stream.write(r"$-\infty")
        elif mode == SPECIAL:
            s: Final[str] = str(text)
            if s not in SPECIAL_CHARS:
                raise ValueError(f"invalid special character: '{s}'")
            stream.write(SPECIAL_CHARS[s])
        else:
            raise ValueError(f"invalid mode {mode} for text '{text}'.")

        if code:
            stream.write("}}")
        if italic:
            stream.write("}}")
        if bold:
            stream.write("}}")

    def __str__(self):
        """
        Get the appropriate file suffix.

        :returns: the file suffix
        :retval 'tex': always
        """
        return "tex"

    @staticmethod
    def instance() -> "LaTeX":
        """
        Get the LaTeX format singleton instance.

        :returns: the singleton instance of the LaTeX format
        """
        attr: Final[str] = "_instance"
        func: Final = LaTeX.instance
        if not hasattr(func, attr):
            setattr(func, attr, LaTeX())
        return getattr(func, attr)



"""
A fitness combining the rank of a solution with the iteration of its creation.

This fitness assignment process is compatible with the simple (mu+lambda)
Evolutionary Algorithm implemented in :class:`~moptipy.algorithms.so.ea.EA`.
It will assign a better fitness to a solution which has a better objective
value. Ties will be broken based on the iteration counter
:attr:`~moptipy.algorithms.so.record.Record.it` of the solution records
:class:`~moptipy.algorithms.so.record.Record`.
"""

from math import inf
from typing import Final

from numpy.random import Generator

from moptipy.algorithms.so.fitness import Fitness, FRecord


class RankAndIteration(Fitness):
    """
    A fitness joining objective rank and creation iteration.

    The fitness assignment strategy will use two pieces of information to
    determine the fitness of a solution:

    1. the rank the solutions by their objective values
       (:attr:`~moptipy.algorithms.so.record.Record.f`). If two solutions have
       the same fitness, they get the same rank, but the next-worst solution
       will then get a rank with is larger by 2.
    2. the iteration index (:attr:`~moptipy.algorithms.so.record.Record.it`)
       relative to the maximum and minimum iteration index.

    It will multiply the rank of the solution with the range of the iteration
    index in the population and then add the maximum iteration index minus the
    iteration index of the solution, i.e.,

    `fitness(x) = rank(f(x)) * (max_it - min_it + 1)  +  (max_it - it(x))`

    This way, better solutions receive better fitness and ties are broken such
    that younger solutions (with higher iteration index) are preferred.
    In combination with best selection
    (:class:`moptipy.algorithms.modules.selections.best.Best`),
    this replicates the behavior of our simple (mu+lambda) Evolutionary
    Algorithm (:class:`~moptipy.algorithms.so.ea.EA`).
    """

    def assign_fitness(self, p: list[FRecord], random: Generator) -> None:
        """
        Assign the rank and iteration fitness.

        :param p: the list of records
        :param random: ignored
        """
        min_it: int = 9_223_372_036_854_775_808  # the minimum iteration index
        max_it: int = -1  # the maximum iteration index

        # In the first iteration, we assign objective value as fitness
        # (for sorting) and get the bounds of the iteration indices.
        for rec in p:
            rec.fitness = rec.f  # set f as fitness for sorting
            it: int = rec.it  # get iteration index from record
            if it < min_it:  # update minimum iteration index
                min_it = it
            if it > max_it:  # update maximum iteration index
                max_it = it

        p.sort()  # sort based on objective values

        it_range: Final[int] = max_it - min_it + 1  # range of it index
        rank: int = -1  # the rank counter
        last_fitness: int | float = -inf
        for i, rec in enumerate(p):
            v = rec.fitness
            if v > last_fitness:  # if fitness differs, step rank
                rank = i + 1  # +1 so smallest-possible rank is 1
                last_fitness = v
            rec.fitness = (rank * it_range) + max_it - rec.it

    def __str__(self):
        """
        Get the name of this default fitness assignment.

        :return: the name of this fitness assignment strategy
        :retval "rankAndIt": always
        """
        return "rankAndIt"
